# Copyright (C) 2021-2026 Arnaud Belcour - Inria, Univ Rennes, CNRS, IRISA Dyliss
# Univ. Grenoble Alpes, Inria, Microcosme
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

import csv
import gzip
import json
import logging
import os
import shutil
import subprocess
import sys
import time
import pandas as pd
import numpy as np
import random

from scipy.optimize import curve_fit
from scipy import __version__ as scipy_version
from Bio import SeqIO
from Bio import __version__ as biopython_version
from shutil import which

from esmecata import __version__ as esmecata_version
from esmecata.utils import is_valid_path, is_valid_dir

logger = logging.getLogger(__name__)


def compute_stat_clustering(output_folder, stat_file=None):
    """Compute stat associated to the number of proteome for each taxonomic affiliations.

    Args:
        output_folder (str): pathname to the result folder containing mmseqs results
        stat_file (str): pathname to the tsv stat file

    Returns:
        clustering_numbers (dict): dict containing observation names (as key) associated with the number of protein clusters
    """
    clustering_taxon_id_file = os.path.join(output_folder, 'proteome_tax_id.tsv')
    result_folder = os.path.join(output_folder, 'computed_threshold')

    proteomes_taxa_id_names = get_proteomes_tax_id_name(clustering_taxon_id_file)

    tax_name_clustering_numbers = {}
    for clustering_file in os.listdir(result_folder):
        clustering_file_path = os.path.join(result_folder, clustering_file)
        with open(clustering_file_path, 'r') as open_clustering_file_path:
            csvreader = csv.reader(open_clustering_file_path, delimiter='\t')
            next(csvreader)
            cluster_0 = []
            cluster_0_5 = []
            cluster_0_95 = []
            for line in csvreader:
                protein_cluster_threshold = float(line[1])
                if protein_cluster_threshold >= 0.95:
                    cluster_0_95.append(line[0])
                if protein_cluster_threshold >= 0.5:
                    cluster_0_5.append(line[0])
                if protein_cluster_threshold >= 0:
                    cluster_0.append(line[0])

        tax_name_clustering_numbers[clustering_file.replace('.tsv', '')] = [cluster_0, cluster_0_5, cluster_0_95]

    clustering_numbers = {}
    for observation_name in proteomes_taxa_id_names:
        tax_name = proteomes_taxa_id_names[observation_name]
        clustering_numbers[observation_name] = tax_name_clustering_numbers[tax_name]

    if stat_file:
        with open(stat_file, 'w') as stat_file_open:
            csvwriter = csv.writer(stat_file_open, delimiter='\t')
            csvwriter.writerow(['observation_name', 'Number_protein_clusters_panproteome', 'Number_protein_clusters_kept', 'Number_protein_clusters_coreproteome'])
            for observation_name in clustering_numbers:
                cluster_0 = len(clustering_numbers[observation_name][0])
                cluster_0_5 = len(clustering_numbers[observation_name][1])
                cluster_0_95 = len(clustering_numbers[observation_name][2])
                csvwriter.writerow([observation_name, cluster_0, cluster_0_5, cluster_0_95])

    return clustering_numbers


def get_proteomes_tax_id_name(proteomes_tax_id_file_path, header_to_extract='tax_id_name'):
    """ Extract tax_name + tax_id associated with observation name.

    Args:
        proteomes_tax_id_file_path (str): pathname to the proteomes_tax_id file.
        header_to_extract (str): name of the header to extract data from (by default tax_id_name).

    Returns:
        proteomes_taxa_id_names (dict): dict containing observation names (as key) associated with tax name + tax_id used for proteomes (as value)
    """
    proteomes_taxa_id_names = {}
    with open(proteomes_tax_id_file_path, 'r') as proteome_tax_file:
        csvreader = csv.DictReader(proteome_tax_file, delimiter='\t')
        for line in csvreader:
            observation_name = line['observation_name']
            data_extract = line[header_to_extract]
            proteomes_taxa_id_names[observation_name] = data_extract

    return proteomes_taxa_id_names


def copy_already_clustered_file(output_folder, already_clustered_obs_name, new_observation_name, file_extension='.tsv'):
    """ Copy files from an observation name with the same tax ID than the new ones.
    Args:
        output_folder (str): path to the ouput result folder
        already_clustered_obs_name (str): name of the already clustered observation name
        new_observation_name (str): name of the new observation name
        file_extension (str): file extension (by default '.tsv')
    """
    already_cluster_output_file = os.path.join(output_folder, already_clustered_obs_name + file_extension)
    cluster_output_file = os.path.join(output_folder, new_observation_name + file_extension)
    shutil.copyfile(already_cluster_output_file, cluster_output_file)


def run_mmseqs(observation_name, observation_name_proteomes, mmseqs_tmp_path, nb_core, mmseqs_options, linclust):
    """Run MMseqs2 on proteomes for an observation name

    Args:
        observation_name (str): observation name associated to a taxonomic affiliations
        observation_name_proteomes (list): list of pathname to each proteomes associated to the observation_name
        mmseqs_tmp_path (str): pathname to the folder which will contain mmseqs results
        nb_core (int): number of CPU-cores for mmseqs
        mmseqs_options (str): options that will be used by mmseqs (default: '--min-seq-id', '0.3', '-c', '0.8')
        linclust (bool): use linclust for faster clustering

    Returns:
        mmseqs_tmp_clustered_tabulated (str): pathname to the mmseqs tabulated file containg protein cluster
        mmseqs_tmp_representative_fasta (str): pathname to the mmseqs fasta file containing representative protein sequences
        mmseqs_consensus_fasta (str): pathname to the mmseqs fasta file containing consensus protein sequences
    """
    mmseqs_tmp_cluster = os.path.join(mmseqs_tmp_path, observation_name)
    is_valid_dir(mmseqs_tmp_cluster)

    mmseqs_tmp_cluster_output = os.path.join(mmseqs_tmp_cluster, 'cluster')
    mmseqs_tmp_clustered_tabulated = mmseqs_tmp_cluster_output+'_cluster.tsv'
    mmseqs_tmp_representative_fasta = mmseqs_tmp_cluster_output + '_rep_seq.fasta'
    mmseqs_consensus_fasta = mmseqs_tmp_cluster_output + '_con_seq.fasta'
    # Run mmmseqs to find protein clusters.

    # Code using mmseqs database and mmseqs modules to cluster instead of easy-cluster.
    mmseqs_tmp_db = os.path.join(mmseqs_tmp_cluster, 'db')
    mmseqs_tmp_db_clustered = os.path.join(mmseqs_tmp_cluster, 'cluster_db')
    mmseqs_tmp_cluster_tmp = os.path.join(mmseqs_tmp_cluster, 'cluster_tmp')
    mmseqs_seq_db =  os.path.join(mmseqs_tmp_cluster, 'cluster_seq')
    mmseqs_profile =  os.path.join(mmseqs_tmp_cluster, 'cluster_profile')
    mmseqs_consensus =  os.path.join(mmseqs_tmp_cluster, 'cluster_consensus')

    if not os.path.exists(mmseqs_tmp_clustered_tabulated):
        # Create database containing the protein sequences from all the proteomes of a taxon.
        subprocess.call(['mmseqs', 'createdb', *observation_name_proteomes, mmseqs_tmp_db, '-v', '2'])

        # Cluster the protein sequences.

        cluster_cmd = ['mmseqs']
        if linclust:
            cluster_cmd += ['linclust']
        else:
            cluster_cmd += ['cluster']

        cluster_cmd += [mmseqs_tmp_db, mmseqs_tmp_db_clustered, mmseqs_tmp_cluster_tmp, '--threads', str(nb_core), '-v', '2']

        # If no option given by the user, use the default options: '--min-seq-id', '0.3', '-c', '0.8'.
        # Sequence identity of 30% and coverage of 80%.
        if not mmseqs_options:
            cluster_cmd += ['--min-seq-id', '0.3', '-c', '0.8']
        else:
            cluster_cmd += mmseqs_options.split(' ')

        subprocess.call(cluster_cmd)

        # Create sequence database with representative from the clustered proteins.
        subprocess.call(['mmseqs', 'createsubdb', mmseqs_tmp_db_clustered, mmseqs_tmp_db, mmseqs_seq_db, '-v', '2'])
        # Create the profile from the clustering.
        subprocess.call(['mmseqs', 'result2profile', mmseqs_seq_db, mmseqs_tmp_db, mmseqs_tmp_db_clustered, mmseqs_profile, '--threads', str(nb_core), '-v', '2'])
        # Create the consensus from the profile.
        subprocess.call(['mmseqs', 'profile2consensus', mmseqs_profile, mmseqs_consensus, '--threads', str(nb_core), '-v', '2'])
        # Create the consensus fasta file.
        subprocess.call(['mmseqs', 'convert2fasta', mmseqs_consensus, mmseqs_consensus_fasta, '-v', '2'])
        # Create TSV resulting files to be analysed after.
        subprocess.call(['mmseqs', 'createtsv', mmseqs_tmp_db, mmseqs_tmp_db, mmseqs_tmp_db_clustered, mmseqs_tmp_clustered_tabulated, '--threads', str(nb_core), '-v', '2'])
        # Create fasta file containing representative proteins (representatives are the first protein of the alignment).
        subprocess.call(['mmseqs', 'convert2fasta', mmseqs_seq_db, mmseqs_tmp_representative_fasta, '-v', '2'])

    # Old clustering method used with easy-cluster.
    """
    if not os.path.exists(mmseqs_tmp_clustered_tabulated):
        # Code using easy-cluster to extract representative protein.
        subprocess.call(['mmseqs', 'easy-cluster', *cluster_fasta_files[cluster], mmseqs_tmp_cluster_output, mmseqs_tmp_cluster, '--threads', str(nb_core), '-v', '2', '--min-seq-id', '0.3'])
    """

    return mmseqs_tmp_clustered_tabulated, mmseqs_tmp_representative_fasta, mmseqs_consensus_fasta


def extrat_protein_cluster_from_mmseqs(mmseqs_tmp_clustered_tabulated, cluster_proteomes_output_file):
    """Extract protein cluster from mmseqs into a dictionary

    Args:
        mmseqs_tmp_clustered_tabulated (str): pathname to the mmseqs tabulated file containg protein cluster
        cluster_proteomes_output_file (str): pathname to the output file showing the content of each protein cluster

    Returns:
        protein_clusters (dict): protein clusters found by mmseqs (representative protein as key and all the protein in the cluster as value)
    """
    # Extract protein clusters in a dictionary.
    # The representative protein is the key and all the proteins in the cluster are the values.
    protein_clusters = {}
    with open(mmseqs_tmp_clustered_tabulated, 'r') as input_file:
        csvreader = csv.reader(input_file, delimiter='\t')
        for row in csvreader:
            if row[0] not in protein_clusters:
                protein_clusters[row[0]] = [row[1]]
            else:
                protein_clusters[row[0]].append(row[1])

    with open(cluster_proteomes_output_file, 'w') as output_file:
        csvwriter = csv.writer(output_file, delimiter='\t')
        for rep_protein in protein_clusters:
            csvwriter.writerow([rep_protein, *[prot for prot in protein_clusters[rep_protein]]])

    return protein_clusters


def compute_proteome_representativeness_ratio(protein_clusters, observation_name_proteomes, computed_threshold_file=None):
    """Compute for each protein cluster the ratio of representation of each proteomes in the cluster.

    Args:
        protein_clusters (dict): protein clusters found by mmseqs (representative protein as key and all the protein in the cluster as value)
        observation_name_proteomes (list): list of pathname to each proteomes associated to the observation_name
        computed_threshold_file (str): pathname to the output file containing the computed ratio

    Returns:
        number_proteomes (int): number of the proteomes for the observation_name
        rep_prot_organims (dict): names of each proteomes associated with each protein cluster
        computed_threshold_cluster (dict): ratio of proteomes representativeness for each protein cluster
    """
    # Retrieve protein ID and the corresponding proteome.
    organism_prots = {}
    for fasta_file in observation_name_proteomes:
        with gzip.open(fasta_file, 'rt') as fasta_handle:
            for record in SeqIO.parse(fasta_handle, 'fasta'):
                compressed_filebasename = os.path.basename(fasta_file)
                fasta_filebasename = os.path.splitext(compressed_filebasename)[0]
                filebasename = os.path.splitext(fasta_filebasename)[0]
                if '|' in record.id:
                    prot_id = record.id.split('|')[1]
                else:
                    prot_id = record.id
                organism_prots[prot_id] = filebasename

    number_proteomes = len(observation_name_proteomes)

    # Compute the ratio between the number of proteome represented by a protein in the cluster and the total number of proteome.
    rep_prot_organims = {}
    computed_threshold_cluster = {}
    for rep_protein in protein_clusters:
        rep_prot_organims[rep_protein] = set([organism_prots[prot] for prot in protein_clusters[rep_protein]])
        computed_threshold_cluster[rep_protein] = len(rep_prot_organims[rep_protein]) / number_proteomes

    if computed_threshold_file:
        # Create a tsv file containing the computer threshold (number of organism in the cluster compared to the total number of organism) for each organisms.
        with open(computed_threshold_file, 'w') as output_file:
            csvwriter = csv.writer(output_file, delimiter='\t')
            csvwriter.writerow(['representative_protein', 'cluster_ratio', 'proteomes'])
            for rep_protein in rep_prot_organims:
                csvwriter.writerow([rep_protein, computed_threshold_cluster[rep_protein], ','.join(rep_prot_organims[rep_protein])])

    return number_proteomes, rep_prot_organims, computed_threshold_cluster


def filter_protein_cluster(protein_clusters, number_proteomes, rep_prot_organims, computed_threshold_cluster,
                           clust_threshold, cluster_proteomes_filtered_output_file=None):
    """Filter protein cluster according to the representation of each proteomes in the cluster.

    Args:
        protein_clusters (dict): protein clusters found by mmseqs (representative protein as key and all the protein in the cluster as value)
        number_proteomes (int): number of the proteomes for the observation_name
        rep_prot_organims (dict): names of each proteomes associated with each protein cluster
        computed_threshold_cluster (dict): ratio of proteomes representativeness for each protein cluster
        clust_threshold (float): threshold to select protein cluster according to the representation of protein proteome in the cluster
        cluster_proteomes_filtered_output_file (str): pathname to the output file showing the protein cluster kept after filtering

    Returns:
        protein_cluster_to_keeps (list): list containing representative protein IDs associated with cluster that are kept
    """
    # Keep a protein cluster according to the ratio and create a list of representative proteins (being the protein cluster to keep).
    protein_cluster_to_keeps = []
    reference_threshold = (clust_threshold * number_proteomes) / number_proteomes

    for rep_protein in rep_prot_organims:
        if computed_threshold_cluster[rep_protein] >= reference_threshold:
            protein_cluster_to_keeps.append(rep_protein)

    if cluster_proteomes_filtered_output_file:
        with open(cluster_proteomes_filtered_output_file, 'w') as filtered_output_file:
            filtered_csvwriter = csv.writer(filtered_output_file, delimiter='\t')
            for rep_protein in protein_cluster_to_keeps:
                filtered_csvwriter.writerow([rep_protein, *[prot for prot in protein_clusters[rep_protein]]])

    # Use set for faster search using 'in'.
    protein_cluster_to_keeps = set(protein_cluster_to_keeps)

    return protein_cluster_to_keeps


def heap_law_curve_fitting(nb_proteomes, nb_new_protein_discovered):
    """ Curve fitting accordng to Heap's Law.

    Args:
        nb_proteomes (list): List of number of genomes
        nb_new_protein_discovered (list): List associated with nb_proteomes list and showing the associated number of newly found genes for the corresponding number of genomes

    Returns:
        k (float): intercept
        alpha (float): constant alpha resulting from the fitting
    """
    # Define Heap's Law: nb_gene_families = k * nb_proteomes^-alpha
    heaps_law = lambda nb_proteomes, k, alpha: k*nb_proteomes**(-alpha)
    heaps_law_func = lambda nb_proteomes, k, alpha: [k*float(nb_proteome)**(-alpha) for nb_proteome in nb_proteomes]
    # Fit Heap's Law with the association between shared protein clusters and proteomes.
    parameter_optimal_values, pcov = curve_fit(f=heaps_law, xdata=nb_proteomes, ydata=nb_new_protein_discovered, p0=[0, 0], bounds=(-np.inf, np.inf),
                                               method='lm')

    k, alpha = parameter_optimal_values

    return k, alpha


def compute_openness_pan_proteomes(esmecata_computed_threshold_folder, output_openness_file, clustering_threhsold=0.5, iteration_nb=1000):
    """ Compute openness of proteomes using Heap's Law.

    Args:
        esmecata_computed_threshold_folder (str): path to esmecata computed threshold folder
        output_openness_file (str): path to output file containing alpha and other data
        clustering_threhsold (float): threshold to select protein cluster according to the representation of protein proteome in the cluster
        iteration_nb (int): number of times permutations will be made on the proteomes list to get different order list of proteomes
    """
    # Define Heap's Law: nb_gene_families = k * nb_proteomes^-alpha
    # Reference used: https://doi.org/10.1016/j.mib.2008.09.006
    # If alpha is superior to 1, pangenome is closed: adding more genomes do not increase number of gene families.
    # If alpha is inferior to 1, pangenome is open: adding more genomes increase the number of gene families.

    proteome_statistics = []
    for computed_threshold_file in os.listdir(esmecata_computed_threshold_folder):
        organism_name = os.path.splitext(computed_threshold_file)[0]
        computed_threshold_filepath = os.path.join(esmecata_computed_threshold_folder, computed_threshold_file)
        df_computed_threshold = pd.read_csv(computed_threshold_filepath, sep='\t', index_col = 0)
        #df_computed_threshold = df_computed_threshold[df_computed_threshold['cluster_ratio']>=clustering_threhsold]
        proteomes_lists = df_computed_threshold['proteomes'].str.split(',')
        # Extract proteomes associated with the organism (and remove the redundancy).
        unique_proteome = list(set([proteome for proteomes in proteomes_lists for proteome in proteomes]))
        # Create several lists of proteome (to avoid computing alpha with only one distribution of proteomes).
        list_proteome_to_iter = [random.sample(unique_proteome, k=len(unique_proteome)) for nb_iter in range(iteration_nb)]
        # Extract proteome and their presence in protein clusters.
        dataset_protein_dict = {proteome: set(df_computed_threshold[df_computed_threshold['proteomes'].str.contains(proteome)].index) for proteome in unique_proteome}

        # Compute mean, median, variance
        proteomes_nb_cluster = [len(dataset_protein_dict[proteome]) for proteome in dataset_protein_dict]
        proteomes_nb_cluster_mean = np.mean(proteomes_nb_cluster)
        proteomes_nb_cluster_median = np.median(proteomes_nb_cluster)
        proteomes_nb_cluster_variance = np.var(proteomes_nb_cluster)
        proteomes_nb_cluster_min = min(proteomes_nb_cluster)
        proteomes_nb_cluster_max = max(proteomes_nb_cluster)
        organism_nb_proteomes = len(unique_proteome)
        nb_protein_clusters_kept = len(df_computed_threshold[df_computed_threshold['cluster_ratio']>=clustering_threhsold].index)

        if organism_nb_proteomes > 1:
            # Compute number of newly found protein clusters when adding proteomes.
            nb_proteomes = []
            nb_new_protein_discovered = []
            for proteome_list in list_proteome_to_iter:
                protein_discovered = set()
                nb_proteome = 0
                for proteome in proteome_list:
                    # Get the protein cluters of the new proteome.
                    protein_clusters_associated = dataset_protein_dict[proteome]
                    nb_proteome += 1
                    if protein_discovered == set():
                        # If it is the first proteome, all its protein clusters correspond to newly found protein clusters.
                        protein_discovered = protein_discovered.union(protein_clusters_associated)
                        new_protein_discovered = protein_discovered
                    else:
                        # Take previously found protein clusters from the other proteomes and extract how many new protein clusters are added by the new proteome.
                        new_protein_discovered = protein_clusters_associated - protein_discovered
                        protein_discovered = protein_discovered.union(protein_clusters_associated)
                    nb_proteomes.append(nb_proteome)
                    nb_new_protein_discovered.append(len(new_protein_discovered))

            k, alpha = heap_law_curve_fitting(nb_proteomes, nb_new_protein_discovered)
        else:
            logger.info('|EsMeCaTa|clustering| Only 1 proteome for {0}, openness not computed for this taxon.'.format(organism_name))
            alpha = 'not computed'

        proteome_statistics.append([organism_name, organism_nb_proteomes, alpha, proteomes_nb_cluster_min, proteomes_nb_cluster_mean, proteomes_nb_cluster_median, proteomes_nb_cluster_max, proteomes_nb_cluster_variance, nb_protein_clusters_kept])

        # Create output dataframe.
        df = pd.DataFrame(proteome_statistics)
        df.columns = ['organism', 'organism_nb_proteomes', 'constant_alpha', 'Minimal number of protein clusters associated with a proteome',
                      'Mean number of protein clusters associated with a proteome', 'Median number of protein clusters associated with a proteome',
                      'Maximal number of protein clusters associated with a proteome', 'Variance of protein clusters associated with a proteome',
                      'Number of protein clusters kept']
        df.to_csv(output_openness_file, index=False, sep='\t')


def make_clustering(proteome_folder, output_folder, nb_core, clust_threshold, mmseqs_options, linclust, remove_tmp):
    """From the proteomes found by esmecata proteomes, create protein cluster for each taxonomic affiliations.

    Args:
        proteome_folder (str): pathname to folder from esmecata folder
        output_folder (str): pathname to the output folder
        nb_core (int): number of CPU-cores to be used by mmseqs
        clust_threshold (float): threshold to select protein cluster according to the representation of protein proteome in the cluster
        mmseqs_options (str): use alternative mmseqs option
        linclust (bool): use linclust
        remove_tmp (bool): remove the tmp files
    """
    starttime = time.time()
    logger.info('|EsMeCaTa|clustering| Begin clustering.')

    # Check if mmseqs is in path.
    mmseqs_path = which('mmseqs')
    if not mmseqs_path:
        logger.critical('|EsMeCaTa|clustering| mmseqs not available in path, esmecata will not be able to cluster the proteomes.')
        sys.exit(1)

    if not is_valid_dir(proteome_folder):
        logger.critical('|EsMeCaTa|clustering| Input must be a folder %s.', proteome_folder)
        sys.exit(1)

    # Use the proteomes folder created by retrieve_proteome.py.
    proteome_tax_id_pathname = os.path.join(proteome_folder, 'proteome_tax_id.tsv')

    if not is_valid_path(proteome_tax_id_pathname):
        logger.critical(f"|EsMeCaTa|clustering| Missing output from esmecata proteomes in {proteome_tax_id_pathname}.")
        sys.exit(1)

    reference_proteins_path = os.path.join(output_folder, 'reference_proteins')
    is_valid_dir(reference_proteins_path)

    cluster_founds_path = os.path.join(output_folder, 'cluster_founds')
    is_valid_dir(cluster_founds_path)

    computed_threshold_path = os.path.join(output_folder, 'computed_threshold')
    is_valid_dir(computed_threshold_path)

    already_performed_clustering = [reference_protein_file.replace('.tsv', '') for reference_protein_file in os.listdir(reference_proteins_path)]

    # Create a dictionary with observation_name as key and the pathname to the proteomes associated to this observation_name as value.
    observation_name_fasta_files = {}
    with open(proteome_tax_id_pathname, 'r') as proteome_tax_file:
        csvreader = csv.DictReader(proteome_tax_file, delimiter='\t')
        for line in csvreader:
            proteomes = line['proteome'].split(',')
            tax_id_name = line['tax_id_name']
            proteomes_path = [os.path.join(proteome_folder, 'proteomes', proteome+'.faa.gz') for proteome in proteomes]
            if tax_id_name not in already_performed_clustering:
                observation_name_fasta_files[tax_id_name] = proteomes_path
            else:
                logger.info('|EsMeCaTa|clustering| Already performed clustering for %s.', tax_id_name)

    is_valid_dir(output_folder)

    # Create metadata file.
    clustering_metadata = {}
    clustering_metadata['tool_options'] = {'proteome_folder': proteome_folder, 'output_folder': output_folder, 'nb_core':nb_core,
                                        'clust_threshold':clust_threshold, 'mmseqs_options': mmseqs_options, 'linclust':linclust,
                                        'remove_tmp': remove_tmp}

    clustering_metadata['tool_dependencies'] = {}
    subprocess_output = subprocess.check_output(['mmseqs', 'version'])
    mmseqs_version = subprocess_output.decode('utf-8')
    clustering_metadata['tool_dependencies']['mmseqs_version'] = mmseqs_version
    clustering_metadata['tool_dependencies']['mmseqs_path'] = mmseqs_path
    clustering_metadata['tool_dependencies']['python_package'] = {}
    clustering_metadata['tool_dependencies']['python_package']['Python_version'] = sys.version
    clustering_metadata['tool_dependencies']['python_package']['biopython'] = biopython_version
    clustering_metadata['tool_dependencies']['python_package']['scipy'] = scipy_version
    clustering_metadata['tool_dependencies']['python_package']['pandas'] = pd.__version__
    clustering_metadata['tool_dependencies']['python_package']['numpy'] = np.__version__
    clustering_metadata['tool_dependencies']['python_package']['esmecata'] = esmecata_version

    # Create tmp folder for mmseqs analysis.
    mmseqs_tmp_path = os.path.join(output_folder, 'mmseqs_tmp')
    is_valid_dir(mmseqs_tmp_path)

    # Create output folder containing shared representative proteins.
    reference_proteins_representative_fasta_path = os.path.join(output_folder, 'reference_proteins_representative_fasta')
    is_valid_dir(reference_proteins_representative_fasta_path)

    reference_proteins_consensus_fasta_path = os.path.join(output_folder, 'reference_proteins_consensus_fasta')
    is_valid_dir(reference_proteins_consensus_fasta_path)

    proteome_taxon_id_file = os.path.join(proteome_folder, 'proteome_tax_id.tsv')
    clustering_taxon_id_file = os.path.join(output_folder, 'proteome_tax_id.tsv')

    if os.path.exists(clustering_taxon_id_file):
        if not os.path.samefile(proteome_taxon_id_file, clustering_taxon_id_file):
            os.remove(clustering_taxon_id_file)
            shutil.copyfile(proteome_taxon_id_file, clustering_taxon_id_file)
    else:
        shutil.copyfile(proteome_taxon_id_file, clustering_taxon_id_file)

    proteomes_taxa_names = get_proteomes_tax_id_name(proteome_taxon_id_file)

    all_tax_names = set(list(proteomes_taxa_names.values())) - set(already_performed_clustering)
    nb_taxa_names = len(all_tax_names)
    logger.info('|EsMeCaTa|clustering| Create consensus proteomes for %d taxa.', nb_taxa_names)

    # For each OTU run mmseqs easy-cluster on them to found the clusters that have a protein in each proteome of the OTU.
    # We take the representative protein of a cluster if the cluster contains a protein from all the proteomes of the OTU.
    # If this condition is not satisfied the cluster will be ignored.
    # Then a fasta file containing all the representative proteins for each OTU is written in representative_fasta folder.
    for index, proteomes_tax_name in enumerate(all_tax_names):
        # Get proteomes associated with taxon name.
        observation_name_proteomes = observation_name_fasta_files[proteomes_tax_name]

        # Change space with '_' to avoid issue.
        proteomes_tax_name = proteomes_tax_name
        # If the computed threshold file exists, mmseqs has already been run.
        mmseqs_tmp_cluster = os.path.join(mmseqs_tmp_path, proteomes_tax_name)
        # Run mmseqs on organism.
        # Delete previous mmseqs2 run if it exists to avoid overwritting issues.
        if os.path.exists(mmseqs_tmp_cluster):
            shutil.rmtree(mmseqs_tmp_cluster)
        mmseqs_tmp_clustered_tabulated, mmseqs_tmp_representative_fasta, mmseqs_consensus_fasta = run_mmseqs(proteomes_tax_name, observation_name_proteomes, mmseqs_tmp_path, nb_core,
                                                                                                             mmseqs_options, linclust)

        # Extract protein clusters from mmseqs results.
        cluster_proteomes_output_file = os.path.join(cluster_founds_path, proteomes_tax_name+'.tsv')
        protein_clusters = extrat_protein_cluster_from_mmseqs(mmseqs_tmp_clustered_tabulated, cluster_proteomes_output_file)

        # Compute proteome representativeness ratio.
        computed_threshold_file = os.path.join(computed_threshold_path, proteomes_tax_name+'.tsv')
        number_proteomes, rep_prot_organims, computed_threshold_cluster = compute_proteome_representativeness_ratio(protein_clusters,
                                                                                                                    observation_name_proteomes, computed_threshold_file)

        # Filter protein cluster for each protein cluster.
        cluster_proteomes_filtered_output_file = os.path.join(reference_proteins_path, proteomes_tax_name+'.tsv')
        protein_cluster_to_keeps = filter_protein_cluster(protein_clusters, number_proteomes, rep_prot_organims, computed_threshold_cluster,
                                                        clust_threshold, cluster_proteomes_filtered_output_file)

        logger.info('|EsMeCaTa|clustering| %d protein clusters kept for %s (%d on %d).', len(protein_cluster_to_keeps), proteomes_tax_name, index+1, nb_taxa_names)

        # Create BioPython records with the representative proteins kept.
        new_records = [record for record in SeqIO.parse(mmseqs_tmp_representative_fasta, 'fasta')
                                        if ('|' in record.id and record.id.split('|')[1] in protein_cluster_to_keeps) or (record.id in protein_cluster_to_keeps)]

        # Do not create fasta file when 0 sequences were kept.
        if len(new_records) > 0:
            # Create output proteome file for OTU.
            representative_fasta_file = os.path.join(reference_proteins_representative_fasta_path, proteomes_tax_name+'.faa')
            SeqIO.write(new_records, representative_fasta_file, 'fasta')
        else:
            logger.info('|EsMeCaTa|clustering| 0 protein clusters %s, no representatitve fasta created (%d on %d).', proteomes_tax_name, index+1, nb_taxa_names)
        del new_records

        # Create BioPython records with the consensus proteins kept.
        consensus_new_records = [record for record in SeqIO.parse(mmseqs_consensus_fasta, 'fasta')
                                        if ('|' in record.id and record.id.split('|')[1] in protein_cluster_to_keeps) or (record.id in protein_cluster_to_keeps)]

        # Do not create fasta file when 0 sequences were kept.
        if len(consensus_new_records) > 0:
            # Create output proteome file for OTU.
            consensus_fasta_file = os.path.join(reference_proteins_consensus_fasta_path, proteomes_tax_name+'.faa')
            SeqIO.write(consensus_new_records, consensus_fasta_file, 'fasta')
        else:
            logger.info('|EsMeCaTa|clustering| 0 protein clusters %s, no consensus fasta created (%d on %d).', proteomes_tax_name, index+1, nb_taxa_names)
        del consensus_new_records

        if remove_tmp:
            shutil.rmtree(mmseqs_tmp_cluster)

    # Compute number of protein clusters kept.
    stat_file = os.path.join(output_folder, 'stat_number_clustering.tsv')
    compute_stat_clustering(output_folder, stat_file)
    # Compute openness of proteomes.
    proteome_openness_file = os.path.join(output_folder, 'stat_openness_proteomes.tsv')
    compute_openness_pan_proteomes(computed_threshold_path, proteome_openness_file, clustering_threhsold=0.5, iteration_nb=100)

    endtime = time.time()
    duration = endtime - starttime
    clustering_metadata['esmecata_clustering_duration'] = duration
    clustering_metadata_file = os.path.join(output_folder, 'esmecata_metadata_clustering.json')
    if os.path.exists(clustering_metadata_file):
        metadata_files = [metadata_file for metadata_file in os.listdir(output_folder) if 'esmecata_metadata_clustering' in metadata_file]
        clustering_metadata_file = os.path.join(output_folder, 'esmecata_metadata_clustering_{0}.json'.format(len(metadata_files)))
        with open(clustering_metadata_file, 'w') as ouput_file:
            json.dump(clustering_metadata, ouput_file, indent=4)
    else:
        with open(clustering_metadata_file, 'w') as ouput_file:
            json.dump(clustering_metadata, ouput_file, indent=4)

    logger.info('|EsMeCaTa|clustering| Clustering complete in {0}s.'.format(duration))



#!/usr/bin/env python

# Copyright (C) 2023-2026 Pauline Hamon-Giraud and Alice Mataigne - Inria, Univ Rennes, CNRS, IRISA Dyliss
# Arnaud Belcour - Univ. Grenoble Alpes, Inria, Microcosme
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

from bokeh.resources import INLINE, CDN
import panel as pn


def create_panel_report(CONFIG, DATA, DATA2, DATA3, fig1, fig4, fig5, fig5b, fig6, fig6b,
                        fig7, fig7b, fig8, fig8b, fig9, fig10, fig10a, fig11,
                        fig12, fig12_details, metadata, output_file, reduce=None):
    """ Create HTML report using panel from plotly figures.
    This function was created as a temporary fix after datapane was no longer maintained.
    Now it is replaced by the function using arakawa report. 

    Args:
        CONFIG (dict): metadata dictionary for plotly.
        DATA (dict): gathers all useful data for the analysis of an esmecata run (inputs, summary tables, taxonomy ...)
        DATA2 (dict): dictionary storing dataframes corresponding to EC number with summary numbers
        DATA3 (dict): dictionary storing dataframes corresponding to GO terms with summary numbers
        fig1 (plotly figure): histogram showing the distribution of proteomes according to taxonomic rank
        fig4 (plotly figure): heatmap showing how many input taxonomic ranks were assigned to the same rank by esmecata and how many were assigned to higher ranks
        fig5 (plotly figure): scatterplot showing the frequency of each EC in taxa
        fig5b (plotly figure): scatterplot showing the frequency of each GO Term in taxa
        fig6 (plotly figure): scatterplot showing the fraction of all EC of the dataset in each taxa
        fig6b (plotly figure): scatterplot showing the fraction of all GO Term of the dataset in each taxa
        fig7 (plotly figure): histogram showing how many EC numbers belongs to each taxa of this rank
        fig7b (plotly figure): histogram showing how many GO Terms belongs to each taxa of this rank
        fig8 (plotly figure): histogram of the fraction of all ECs of the dataset in each taxa
        fig8b (plotly figure): histogram of the fraction of all GO Terms of the dataset in each taxa
        fig9 (plotly figure): sunburst representing the taxonomic variety of taxa in the dataset.
        fig10 (plotly figure): sankey graph showing compression and taxonomic ranks rearrangement between input and output taxa of an Esmecata run
        fig10a (plotly figure): barplot showing the number proteomes according to the taxonomic rank
        fig11 (plotly figure): sunburst summarizing EC numbers categories, counts and proportions for the whole dataset
        fig12 (plotly figure): scatterplot showing the representativeness ratio and the number of associated protein clusters according to the taxonomic rank
        fig12_details (plotly figure): scatterplot showing details of the representativeness ratio and the number of associated protein clusters according to each cluster
        metadata (dict): all metadata of the workflow (made by merging esmecata metadata json from proteomes, clustering and annotation)
        output_file (str): path to HTML output file
    """
    styles = {
        'background-color': '#F6F6F6', 'border': '2px solid black',
        'border-radius': '5px', 'padding': '10px'
    }

    page_one = pn.Column(
        pn.pane.Markdown("""# Reduction of taxonomic diversity between EsMeCaTa's inputs and outputs
                    The sunburst below displays the taxonomic diversity of inputs, i.e. all the taxonomic affiliations listed in the tsv input file. White cells indicates inputs whose ranks were not retained by EsMeCaTa, which went up to the superior taxonomic rank to find proteomes.""",
                    styles=styles, renderer='markdown'),
        pn.Column('## Sunburst of taxon selected by EsMeCaTa',
            pn.pane.Plotly(fig9.update_layout(autosize=True)),
            sizing_mode='scale_both'),

        pn.pane.Markdown("""# \"Compression\" of taxonomic diversity between inputs and outputs
                    The sankey diagram below also displays the initial taxonomic diversity of taxonomic affiliations given to EsMeCaTa as inputs. The first column details the content of the input file : one block is equivalent to one taxonomic affiliation, and the block height indicates the number of times this affiliation appears in the input file. The second column details which of these taxonomic affiliations were merged together (i.e. 'compressed') because their taxonomyies were identical in the input file. Finally, The first column displays the ranks EsMeCaTa attributed to each of them : at this step, some taxonomic affiliations can be merged because EsMeCaTa went up the taxonomic ranks to find proteomes.""", styles=styles),
        pn.Column('## Sankey diagram of taxonomic compression',
            pn.pane.Plotly(fig10.update_layout(autosize=True)),
            sizing_mode='scale_both'),
    )
    page_two = pn.Column(
        pn.pane.Markdown("""# Proteomes summary
                     When proteomes of the lowest taxonomic rank of an input taxa are not found, EsMeCaTa goes up in the taxonomic levels to found suited proteomes. In complement to the figures in the first panel, the heatmap below details the difference between the lowest taxonomic rank of all input taxonomic observations (y-axis) and the lowest rank atributed by EsMeCaTa (x-axis).""", styles=styles),
        pn.Column('## Input and output ranks difference',
            pn.pane.Plotly(fig4.update_layout(modebar=CONFIG, autosize=True)),
            sizing_mode='scale_both'),
        pn.pane.HTML("""<p>The barplot belows details how many proteomes were assigned to inputs taxonomic affiliations (grouped and colored by phylum). The number of proteomes found can vary from only a few to dozens or hundreds, depending on the data available on UniProt for each taxonomic group.</p>""", styles=styles),
        pn.Column('## Distribution of the number of proteomes found',
            pn.pane.Plotly(fig1.update_layout(modebar=CONFIG, autosize=True)),
            sizing_mode='scale_both'),
        pn.Column('## Distribution of proteomes according to taxonomic ranks',
            pn.pane.Plotly(fig10a.update_layout(modebar=CONFIG, autosize=True)),
            sizing_mode='scale_both'),
    )
    page_three = pn.Column(
            pn.pane.HTML("""<h2>Clustering summary</h2>
                         <p>In the clustering step, for the set of proteomes associated to each input taxa, EsMeCaTa searches for clusters of similar proteic sequences among these proteomes. This step is supervised by a threshold between 0 and 1, 1 meaning that a cluster contains sequences from each proteomes, 0.5 meaning that a cluster contains sequences from at least a half of the proteomes (etc). This threshold can then be an approximation of the pan-proteome (0) and the core-proteome (1). Clusters below the threshold are discarded. The figure below displays how many clusters were retained at each threshold value. The vertical red line displays the threshold set dor this run of EsMeCata.</p>""", styles=styles),
        pn.Column('## Proteomes representativeness',
            pn.pane.Plotly(fig12.update_layout(modebar=CONFIG, autosize=True)),
            sizing_mode='scale_both'),
        pn.Column('## Proteomes representativeness details',
            pn.pane.Plotly(fig12_details.update_layout(modebar=CONFIG, autosize=True)),
            sizing_mode='scale_both'),
    )
    page_four = pn.Column(
        # EC numbers figures
        pn.FlexBox(
            pn.pane.HTML("""<h2>EC numbers frequencies among taxa</h2>
                        <p>Numbers and figures below detail how EC numbers vary among taxa, i.e. if an EC number is found in many taxa (a generic EC number) or only in very few taxa (a specific EC number).</p>""", styles=styles),
            pn.indicators.Number(name='EC found in more than 90% of taxa', value=DATA2["n_9_an"], format='{value}'),
            pn.indicators.Number(name='EC found in less than 10% of taxa', value=DATA2["n_1_an"], format='{value}'),
            pn.indicators.Number(name='EC found between 10% and 90% of taxa', value=DATA2["n91_an"], format='{value}'),
        ),
        pn.Column(
            pn.Column('## EC numbers frequencies in observations',
                pn.pane.Plotly(fig5.update_layout(modebar=CONFIG, autosize=True)),
                sizing_mode='scale_both'),
            pn.Column('## EC numbers frequencies in observations (barplot version)',
                pn.pane.Plotly(fig7.update_layout(modebar=CONFIG, autosize=True)),
                sizing_mode='scale_both'),
        ),

        pn.FlexBox(
            pn.pane.HTML("""<h2>Taxa content in EC numbers</h2>
                        <p>Numbers and figures below detail which taxa are annotated with most of the EC numbers of the whole dataset or only a few EC numbers. For instance, taxa with many EC numbers could have bigger genomes or be generalist species.</p>""", styles=styles),
            pn.indicators.Number(name="Taxa with more than 90% of all the dataset's EC numbers", value=DATA2["n_9_ob"], format='{value}'),
            pn.indicators.Number(name="Taxa with less than 10% of all the dataset's EC numbers", value=DATA2["n_1_ob"], format='{value}'),
            pn.indicators.Number(name="Taxa with between 10% and 90% of all the dataset's EC numbers", value=DATA2["n91_ob"], format='{value}'),
        ),
        pn.Column(
            pn.Column('## Taxa content in EC numbers',
                pn.pane.Plotly(fig6.update_layout(modebar=CONFIG, autosize=True)),
                sizing_mode='scale_both'),
            pn.Column('## Taxa content in EC numbers (barplot version)',
                pn.pane.Plotly(fig8.update_layout(modebar=CONFIG, autosize=True)),
                sizing_mode='scale_both'),
        ),

        pn.pane.HTML("<h2>EC numbers classes, counts and proportions</h2>", styles=styles),
        pn.Column('## EC numbers classes, counts and proportions',
            pn.pane.Plotly(fig11.update_layout(modebar=CONFIG, autosize=True)),
            sizing_mode='scale_both'),

        # GO terms figures
        pn.FlexBox(
            pn.pane.HTML("""<h2>GO terms frequencies among taxa</h2>
                        <p>Numbers and figures below detail how GO terms vary among taxa, i.e. if a GO term is found in many taxa (a generic GO term) or only in very few taxa (a specific GO term).</p>""", styles=styles),
            pn.indicators.Number(name="GO terms found in more than 90% of taxa", value=DATA3["n_9_an"], format='{value}'),
            pn.indicators.Number(name="GO terms found in less than 10% of taxa", value=DATA3["n_1_an"], format='{value}'),
            pn.indicators.Number(name="GO terms found between 10% and 90% of taxa", value=DATA3["n91_an"], format='{value}'),
        ),
        pn.Column(
            pn.Column('## GO terms frequencies in taxa',
                pn.pane.Plotly(fig5b.update_layout(modebar=CONFIG, autosize=True)),
                sizing_mode='scale_both'),
            pn.Column('## GO terms frequencies in taxa (barplot version)',
                pn.pane.Plotly(fig7b.update_layout(modebar=CONFIG, autosize=True)),
                sizing_mode='scale_both'),
        ),

        pn.FlexBox(
        pn.pane.HTML("""<h2>Taxa content in GO terms</h2>
                     <p>Numbers and figures below detail which taxa are annotated with most of the GO terms of the whole dataset or only a few Go terms. For instance, taxa with many Go terms could have bigger genomes or be generalist species.</p>""", styles=styles),
            pn.indicators.Number(name="Taxa with more than 90% of all the dataset's GO terms", value=DATA3["n_9_ob"], format='{value}'),
            pn.indicators.Number(name="Taxa with less than 10% of all the dataset's GO terms", value=DATA3["n_1_ob"], format='{value}'),
            pn.indicators.Number(name="Taxa with between 10% and 90% of all the dataset's GO terms", value=DATA3["n91_ob"], format='{value}'),
        ),
        pn.Column(
            pn.Column("## Observations' content in GO terms",
                pn.pane.Plotly(fig6b.update_layout(modebar=CONFIG, autosize=True)),
                sizing_mode='scale_both'),
            pn.Column("## Observations' content in GO terms (barplot version)",
                pn.pane.Plotly(fig8b.update_layout(modebar=CONFIG, autosize=True)),
                sizing_mode='scale_both'),
        ),
        height=5000
    )

    if not DATA["DISCARDED"].empty:    
        discarded_panel_content_panel = pn.pane.DataFrame(DATA["DISCARDED"])
    else:
        discarded_panel_content_panel =  pn.pane.HTML("<p>None of input the taxonomic observations were discarded</p>", styles=styles)

    page_five = pn.Column(
                pn.FlexBox(
                    pn.indicators.Number(name="Number of inputs", value=DATA["N_IN"], format='{value}'),
                    pn.indicators.Number(name="Kept by EsMeCaTa", value=DATA["N_OUT"], format='{value}'),
                    pn.indicators.Number(name="Discarded", value=DATA["N_DISCARDED"], format='{value}'),
                ),
                pn.FlexBox(
                    pn.pane.HTML("<h2>Output stats</h2><p>Taxonomic ranks are NCBI ranks returned by ete4</p>", styles=styles),
                    pn.pane.DataFrame(DATA["DF_STATS"]),
                ),
                pn.FlexBox(
                    pn.pane.HTML("<h2>Discarded</h2><p>Taxonomic ranks were not inferred; only names are displayed</p>", styles=styles),
                    discarded_panel_content_panel
                ),
    )

    page_six = pn.Column(
        pn.pane.JSON(metadata, name="Metadata", depth=-1),
    )

    tabs = pn.Tabs(("Inputs and outputs ranks comparison", page_one),
                   ("Proteomes summary", page_two),
                   ("Clustering summary", page_three),
                   ("Annotation summary", page_four),
                   ("Data summary", page_five),
                   ("Metadata", page_six),
                   height_policy="fit", sizing_mode="fixed")

    panel = pn.template.VanillaTemplate(
        title="EsMeCaTa report",
        main=[tabs],
        header_color='black',
        header_background='white'
    )
 
    if reduce is True:
        panel.save(output_file, resources=CDN)
    else:
        panel.save(output_file, resources=INLINE)




# Copyright (C) 2023-2026 Alice Mataigne and Pauline Hamon-Giraud - Inria, Univ Rennes, CNRS, IRISA Dyliss
# Arnaud Belcour - Univ. Grenoble Alpes, Inria, Microcosme
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

import csv
import os.path

import plotly.graph_objects as go
from ete4 import NCBITaxa
from typing import Dict, List, Set, Tuple

from esmecata.report.esmecata_compression import RANK2COL

# CONSTANTS ==========================================================================================================

NCBI = NCBITaxa()


# FUNCTIONS ==========================================================================================================
def get_fig_parameters(tax_diff_file: str) -> Dict[str, List]:
    """ Fill a dictionary with input parameters for the sunburst figure creation.

    Parameters
    ----------
    tax_diff_file: str
        Path of "taxonomy_diff.tsv" file from proteome step.

    Returns
    -------
    Dict[str, List]
        Dictionary of parameters for sunburst figure. Multiple lists of size N with following information for each
         taxon i in [1:n]:
            - ID: tax ID
            - Name: taxon name
            - Rank: taxonomic rank
            - Parent: tax ID of the parent taxon
            - Count Input: Number of OTU corresponding to the taxon in input data
            - Count Esmecata: Number of OTU corresponding to the taxon after esmecata proteome selection
            - Color: Hexadecimal color value corresponding of the taxonomic rank
            - Shape: Shape of the sunburst cell, hatched if the taxon is not kept after esmecata filter, plain otherwise
    """
    count_input = dict()
    count_esmecata = dict()
    tax_names = dict()
    tax_ranks = dict()
    parent = dict()
    with open(tax_diff_file, 'r') as f:
        reader = csv.reader(f, delimiter='\t')
        reader.__next__()
        for line in reader:
            count, input_tax_id, esmecata_tax_id = len(line[6].split(';')), int(line[4]), int(line[5])
            lineage_ids = NCBI.get_lineage(input_tax_id)
            count_input = fill_count_input(count_input, lineage_ids, count)
            tax_names = fill_tax_names(tax_names, lineage_ids)
            parent = fill_parent(parent, lineage_ids)
            tax_ranks = fill_tax_ranks(tax_ranks, lineage_ids)
            count_esmecata = fill_count_esmecata(count_esmecata, count, esmecata_tax_id)

    to_exclude, parent = select_root(count_input, parent)
    return fill_parameters(tax_names, tax_ranks, parent, count_input, count_esmecata, to_exclude)


def fill_count_input(count_input: Dict[int, int], lineage_ids: List[int], count: int) -> Dict[int, int]:
    """ Fill the count_input dictionary for Count input fig parameter for one taxon.

    Parameters
    ----------
    count_input: Dict[int, int]
        Dictionary associating to each tax id, the number of OTU associated to the taxon.
    lineage_ids: List[int]
        List of tax id representing the lineage of the taxon (root to taxon order)
    count: int
        Number of OTU corresponding to the taxon.

    Returns
    -------
    Dict[int, int]
        Dictionary associating to each tax id, the number of OTU associated to the taxon.
    """
    for tax_id in lineage_ids:
        if tax_id not in count_input.keys():
            count_input[tax_id] = count
        else:
            count_input[tax_id] += count
    return count_input


def fill_tax_names(tax_names: Dict[int, str], lineage_ids: List[int]) -> Dict[int, str]:
    """ Fill the tax_names dictionary for Name fig parameter for one taxon.

    Parameters
    ----------
    tax_names: Dict[int, str]
        Dictionary associating to each tax id, its taxon name.
    lineage_ids: List[int]
        List of tax id representing the lineage of the taxon (root to taxon order)

    Returns
    -------
    Dict[int, str]
        Dictionary associating to each tax id, its taxon name.
    """
    lineage_names = NCBI.translate_to_names(lineage_ids)
    for i in range(len(lineage_ids)):
        tax_names[lineage_ids[i]] = lineage_names[i]
    return tax_names


def fill_parent(parent: Dict[int, int or str], lineage_ids: List[int]) -> Dict[int, int or str]:
    """ Fill the parent dictionary for Parent fig parameter for one taxon.

    Parameters
    ----------
    parent: Dict[int, int or str]
        Dictionary associating to each tax id, its parent taxon ID.
    lineage_ids: List[int]
        List of tax id representing the lineage of the taxon (root to taxon order)

    Returns
    -------
    Dict[int, int or str]
        Dictionary associating to each tax id, its parent taxon ID ('' if root).
    """
    for i in range(len(lineage_ids)):
        if lineage_ids[i] != 1:
            parent[lineage_ids[i]] = lineage_ids[i - 1]
        else:
            parent[lineage_ids[i]] = ''
    return parent


def fill_tax_ranks(tax_ranks: Dict[int, str], lineage_ids: List[int]) -> Dict[int, str]:
    """ Fill the tax_ranks dictionary for Rank fig parameter for one taxon.

    Parameters
    ----------
    tax_ranks: Dict[int, str]
        Dictionary associating to each tax id, its taxonomic rank.
    lineage_ids: List[int]
        List of tax id representing the lineage of the taxon (root to taxon order)

    Returns
    -------
    Dict[int, str]
        Dictionary associating to each tax id, its taxonomic rank.
    """
    lineage_ranks = NCBI.get_rank(lineage_ids)
    return {**tax_ranks, **lineage_ranks}


def fill_count_esmecata(count_esmecata: Dict[int, int], count: int, esmecata_id: int) -> Dict[int, int]:
    """ Fill the count_esmecata dictionary for "Count esmecata" fig parameter for one esmecata taxon kept.

    Parameters
    ----------
    count_esmecata: Dict[int, int]
        Dictionary associating to each tax id, the number of OTU associated to the taxon for taxon kept.
    count: int
        Number of OTU corresponding to the taxon.
    esmecata_id
        Tax ID of the esmecata taxon kept.
    Returns
    -------
    Dict[int, int]
        Dictionary associating to each tax id, the number of OTU associated to the taxon for taxon kept.

    """
    lineage_ids = NCBI.get_lineage(esmecata_id)
    for tax_id in lineage_ids:
        if tax_id not in count_esmecata.keys():
            count_esmecata[tax_id] = count
        else:
            count_esmecata[tax_id] += count
    return count_esmecata


def select_root(count_input: Dict[int, int], parent: Dict[int, int or str]) -> Tuple[Set[int], Dict[int, int]]:
    """ Select the root to use and return tax IDs to exclude from the figure. Keep as root the tax ID shared by all
    taxa having the lowest taxonomic rank.

    Parameters
    ----------
    count_input: Dict[int, int]
        Dictionary associating to each tax id, the number of OTU associated to the taxon.
    parent: Dict[int, int or str]
        Dictionary associating to each tax id, its parent taxon ID.

    Returns
    -------
    Tuple[Set[int], Dict[int, int]]
        Set of tax IDs to exclude and Parent dictionary modified for the root ID.
    """
    max_count = max(count_input.values())
    roots = [x for x, y in count_input.items() if y == max_count]
    root = 1
    for ch_id in parent.values():
        if ch_id != '':
            if parent[ch_id] in roots and ch_id not in roots:
                root = parent[ch_id]
    parent[root] = ''

    return set(roots).difference({root}), parent


def fill_parameters(tax_names: Dict[int, str], tax_ranks: Dict[int, str], parent: Dict[int, int or str],
                    count_input: Dict[int, int], count_esmecata: Dict[int, int], to_exclude: Set[int]) \
        -> Dict[str, List]:
    """ Fill a dictionary with parameters for sunburst figure.

    Parameters
    ----------
    tax_names: Dict[int, str]
        Dictionary associating to each tax id, its taxon name.
    tax_ranks: Dict[int, str]
        Dictionary associating to each tax id, its taxonomic rank.
    parent: Dict[int, int or str]
        Dictionary associating to each tax id, its parent taxon ID.
    count_input: Dict[int, int]
        Dictionary associating to each tax id, the number of OTU associated to the taxon.
    count_esmecata: Dict[int, int]
        Dictionary associating to each tax id, the number of OTU associated to the taxon for taxon kept.
    to_exclude: Set[int]
        Set of tax IDs to exclude for root selection.

    Returns
    -------
    Dict[str, List]:
        Dictionary of parameters for sunburst figure. Multiple lists of size N with following information for each
        taxon i in [1:n]:
            - ID: tax ID
            - Name: taxon name
            - Rank: taxonomic rank
            - Parent: tax ID of the parent taxon
            - Count Input: Number of OTU corresponding to the taxon in input data
            - Count Esmecata: Number of OTU corresponding to the taxon after esmecata proteome selection
            - Color: Hexadecimal color value corresponding of the taxonomic rank
            - Shape: Shape of the sunburst cell, hatched if the taxon is not kept after esmecata filter, plain otherwise
    """
    data = {'ID': list(),
            'Name': list(),
            'Rank': list(),
            'Parent': list(),
            'Count Input': list(),
            'Count Esmecata': list(),
            'Color': list(),
            'Shape': list()}
    for tax_id in tax_names.keys():
        data['ID'].append(tax_id)
        data['Name'].append(tax_names[tax_id])
        data['Rank'].append(tax_ranks[tax_id])
        data['Parent'].append(parent[tax_id])
        data['Count Input'].append(count_input[tax_id])
        # data['Color'].append(RANK2COL[tax_ranks[tax_id]])
        if tax_id in count_esmecata.keys():
            data['Count Esmecata'].append(count_esmecata[tax_id])
            data['Color'].append(RANK2COL[tax_ranks[tax_id]])
            #data['Shape'].append('')
        else:
            data['Count Esmecata'].append(0)
            data['Color'].append('#ffffff')
            #data['Shape'].append('/')

    return data


def generate_sunburst_fig(data: Dict[str, List[str or int]], output: str):
    """ Generate a Sunburst figure and save it to output path.

    Parameters
    ----------
    data: Dict[str, List]
        Dictionary of parameters for sunburst figure. Multiple lists of size N with following information for each
         taxon i in [1:n]:
            - ID: tax ID
            - Name: taxon name
            - Rank: taxonomic rank
            - Parent: tax ID of the parent taxon
            - Count Input: Number of OTU corresponding to the taxon in input data
            - Count Esmecata: Number of OTU corresponding to the taxon after esmecata proteome selection
            - Color: Hexadecimal color value corresponding of the taxonomic rank
            - Shape: Shape of the sunburst cell, hatched if the taxon is not kept after esmecata filter, plain otherwise
    output: str
        Path to output to save the figure without extension
    """
    # Create Sunburst
    fig = go.Figure(go.Sunburst(labels=data['Name'], parents=data['Parent'], values=data['Count Input'], ids=data['ID'],
                                branchvalues='total', hovertext=[f'Rank: {x}' for x in data['Rank']], opacity = 0.8,
                                hoverinfo='label+value+percent entry+text', maxdepth=10,
                                insidetextfont=dict(color='#000000', size=16), #cccccc
                                marker=dict(colors=data['Color'],
                                            line=dict(color='#000000'),                                            
                                            pattern=dict(shape=data['Shape'])
                                            )
                                )
                    )

    # Add ranks color annotation
    m = len(set(data['Rank']))
    i = 0
    for rank, color in RANK2COL.items():
        if rank in data['Rank']:
            i += 1
            fig.add_annotation(dict(font=dict(color='#ffffff', size=16), x=0, y=1 - (i / m), showarrow=False, text=rank,
                                    textangle=0, xanchor='left', xref="paper", yref="paper",
                                    bgcolor=color,
                                    opacity=0.8))

    # Update html page layout
    fig.update_layout(paper_bgcolor="#ffffff", #1b1b1b
                      font_color='#ffffff',
                      # width=2000,
                      height=1300,
                      font=dict(family="Courier New, monospace",size=14,color="black"),
                      margin=go.layout.Margin(
                        l=0, #left margin
                        r=0, #right margin
                        b=0, #bottom margin
                        t=0) #top margin
    )
    fig.write_html(f'{output}.html')

    return fig


# MAIN FUNCTION =====================================================================================================

def esmecata2taxonomy(run_name: str, output: str):
    """ Generate a Sunburst representing the taxonomic variety of OTU of a run. Compare ranks from input data to ranks
    kept by the proteome selection.

    Parameters
    ----------
    run_name: str
    output: str
    """
    tax_comp_file = os.path.join(run_name, '0_proteomes', 'taxonomy_diff.tsv')
    data = get_fig_parameters(tax_comp_file)
    fig = generate_sunburst_fig(data, output)
    
    return fig



import shutil
import subprocess
import json
import os
import csv
from esmecata.core.eggnog import annotate_with_eggnog

EXPECTED_RESULTS = {
    'Cluster_1': {'proteomes': 18, 'protein_clusters': 573, 'GOs': 2557, 'ECs': 367}
}

def fake_creation_eggnog(input_path, taxon_name, output_dir, temporary_dir, eggnog_database_path, nb_core, no_dbmem=False, taxon_id_scope=None):
    output_folder = 'output_folder'
    annotation_expected = os.path.join('test_data', 'annotation_expected')

    eggnog_folder_expected = os.path.join(annotation_expected, 'eggnog_output')
    for expected_file in os.listdir(eggnog_folder_expected):
        shutil.copyfile(os.path.join(eggnog_folder_expected, expected_file), os.path.join(output_folder, 'eggnog_output', expected_file))
 
    eggnog_fasta_expected = os.path.join(annotation_expected, 'merge_fasta', '2.faa')
    ouput_fasta_expected = os.path.join(output_folder, 'merge_fasta', '2.faa')
    shutil.copyfile(eggnog_fasta_expected, ouput_fasta_expected)


def test_annotate_with_eggnog_mocked(mocker):
    eggnog_version = "2.1.9"
    # Mock call to eggnog-mapper by creating expected files.
    mocker.patch("esmecata.core.eggnog.get_eggnog_version", return_value=eggnog_version)
    mocker.patch("esmecata.core.eggnog.call_to_emapper", wraps=fake_creation_eggnog)
    annotation_input = os.path.join('test_data', 'annotation_input')
    annotate_with_eggnog(annotation_input, 'output_folder', 'eggnog_database', 1)

    results = {}
    annotation_stat_file = os.path.join('output_folder', 'stat_number_annotation.tsv')
    with open(annotation_stat_file, 'r') as open_annotation_stat_file:
        csvreader = csv.DictReader(open_annotation_stat_file, delimiter='\t')
        for line in csvreader:
            observation_name = line['observation_name']
            results[observation_name] = {}
            results[observation_name]['GOs'] = int(line['Number_go_terms'])
            results[observation_name]['ECs'] = int(line['Number_ecs'])

    for observation_name in EXPECTED_RESULTS:
        for data in results[observation_name]:
            assert results[observation_name][data] == EXPECTED_RESULTS[observation_name][data]

    eggnog_metadata_file = os.path.join('output_folder', 'esmecata_metadata_annotation.json')
    with open(eggnog_metadata_file, 'r') as open_json_file:
        json_data = json.load(open_json_file)
        assert json_data['tool_options']['tool_dependencies']['eggnog_mapper'] == eggnog_version
    shutil.rmtree('output_folder')



import pandas as pd
import os
import shutil
import subprocess

from esmecata.gseapy.gseapy_orsum import taxon_rank_annotation_enrichment

ANNOTATION_NAMES_JSON = os.path.join('test_data', 'annotation_names.json')


def test_taxon_rank_annotation_enrichment_selected():
    input_folder = os.path.join('test_data', 'annotation_output')
    output_folder = 'output_folder'
    grouping = 'selected'
    taxa_list_file = os.path.join('test_data', 'taxa_list.tsv')
    expected_terms = ['1.1.1.86', '2.8.4.1']
    if not os.path.exists(output_folder):
        os.mkdir(output_folder)
    taxon_rank_annotation_enrichment(input_folder, output_folder, grouping, taxa_lists_file=taxa_list_file, annot_names_file=ANNOTATION_NAMES_JSON, orsum_minterm_size=4)

    result_file = os.path.join(output_folder, 'orsum_output_folder', 'filteredResult-Summary.tsv')
    df = pd.read_csv(result_file, sep='\t')
    representing_terms = df['Representing term id']

    assert sorted(representing_terms) == sorted(expected_terms)
    shutil.rmtree(output_folder)


def test_taxon_rank_annotation_enrichment_selected_download():
    input_folder = os.path.join('test_data', 'annotation_output')
    output_folder = 'output_folder'
    grouping = 'selected'
    taxa_list_file = os.path.join('test_data', 'taxa_list.tsv')
    expected_terms = ['1.1.1.86', '2.8.4.1']
    if not os.path.exists(output_folder):
        os.mkdir(output_folder)
    taxon_rank_annotation_enrichment(input_folder, output_folder, grouping, taxa_lists_file=taxa_list_file, annot_names_file='download', orsum_minterm_size=4)

    result_file = os.path.join(output_folder, 'orsum_output_folder', 'filteredResult-Summary.tsv')
    df = pd.read_csv(result_file, sep='\t')
    representing_terms = df['Representing term id']

    assert sorted(representing_terms) == sorted(expected_terms)
    shutil.rmtree(output_folder)


def test_taxon_rank_annotation_enrichment_selected_cli():
    input_folder = os.path.join('test_data', 'annotation_output')
    output_folder = 'output_folder'
    grouping = 'selected'
    taxa_list_file = os.path.join('test_data', 'taxa_list.tsv')
    expected_terms = ['1.1.1.86', '2.8.4.1']
    if not os.path.exists(output_folder):
        os.mkdir(output_folder)
    subprocess.call(['esmecata_gseapy', 'gseapy_enrichr', '-f', input_folder, '-o', output_folder, '--grouping', grouping, '--taxa-list', taxa_list_file, '--orsumMinTermSize', '4',
                     '--annot-names', ANNOTATION_NAMES_JSON])

    result_file = os.path.join(output_folder, 'orsum_output_folder', 'filteredResult-Summary.tsv')
    df = pd.read_csv(result_file, sep='\t')
    representing_terms = df['Representing term id']

    assert sorted(representing_terms) == sorted(expected_terms)
    shutil.rmtree(output_folder)


def test_taxon_rank_annotation_enrichment_selected_cutoff():
    input_folder = os.path.join('test_data', 'annotation_output')
    output_folder = 'output_folder'
    grouping = 'selected'
    taxa_list_file = os.path.join('test_data', 'taxa_list.tsv')
    expected_terms = ['1.1.1.86', '2.8.4.1']
    if not os.path.exists(output_folder):
        os.mkdir(output_folder)
    taxon_rank_annotation_enrichment(input_folder, output_folder, grouping, taxa_lists_file=taxa_list_file, annot_names_file=ANNOTATION_NAMES_JSON, orsum_minterm_size=4, selected_adjust_pvalue_cutoff=0.1)

    result_file = os.path.join(output_folder, 'orsum_output_folder', 'filteredResult-Summary.tsv')
    df = pd.read_csv(result_file, sep='\t')
    representing_terms = df['Representing term id']

    assert sorted(representing_terms) == sorted(expected_terms)
    shutil.rmtree(output_folder)


def test_taxon_rank_annotation_enrichment_selected_cutoff_cli():
    input_folder = os.path.join('test_data', 'annotation_output')
    output_folder = 'output_folder'
    grouping = 'selected'
    taxa_list_file = os.path.join('test_data', 'taxa_list.tsv')
    expected_terms = ['1.1.1.86', '2.8.4.1']
    if not os.path.exists(output_folder):
        os.mkdir(output_folder)
    subprocess.call(['esmecata_gseapy', 'gseapy_enrichr', '-f', input_folder, '-o', output_folder, '--grouping', grouping, '--taxa-list', taxa_list_file, '--annot-names', ANNOTATION_NAMES_JSON,
                     '--orsumMinTermSize', '4', '--gseapyCutOff', '0.1'])

    result_file = os.path.join(output_folder, 'orsum_output_folder', 'filteredResult-Summary.tsv')
    df = pd.read_csv(result_file, sep='\t')
    representing_terms = df['Representing term id']

    assert sorted(representing_terms) == sorted(expected_terms)
    shutil.rmtree(output_folder)


def test_taxon_rank_annotation_enrichment_tax_rank():
    input_folder = os.path.join('test_data', 'annotation_output')
    output_folder = 'output_folder'
    grouping = 'tax_rank'
    expected_terms = ['1.1.1.86', '2.8.4.1']
    if not os.path.exists(output_folder):
        os.mkdir(output_folder)
    taxon_rank_annotation_enrichment(input_folder, output_folder, grouping, annot_names_file=ANNOTATION_NAMES_JSON, orsum_minterm_size=4)

    result_file = os.path.join(output_folder, 'orsum_output_folder', 'filteredResult-Summary.tsv')
    df = pd.read_csv(result_file, sep='\t')
    representing_terms = df['Representing term id']

    assert sorted(representing_terms) == sorted(expected_terms)
    shutil.rmtree(output_folder)


def test_taxon_rank_annotation_enrichment_tax_rank_cli():
    input_folder = os.path.join('test_data', 'annotation_output')
    output_folder = 'output_folder'
    grouping = 'tax_rank'
    expected_terms = ['1.1.1.86', '2.8.4.1']
    if not os.path.exists(output_folder):
        os.mkdir(output_folder)
    subprocess.call(['esmecata_gseapy', 'gseapy_enrichr', '-f', input_folder, '-o', output_folder, '--grouping', grouping,
                     '--annot-names', ANNOTATION_NAMES_JSON, '--orsumMinTermSize', '4'])

    result_file = os.path.join(output_folder, 'orsum_output_folder', 'filteredResult-Summary.tsv')
    df = pd.read_csv(result_file, sep='\t')
    representing_terms = df['Representing term id']

    assert sorted(representing_terms) == sorted(expected_terms)
    shutil.rmtree(output_folder)


def test_taxon_rank_annotation_enrichment_tax_rank_cli_input_file():
    input_folder = os.path.join('test_data', 'annotation_output', 'dataset_annotation.tsv')
    taxon_id_file = os.path.join('test_data', 'annotation_output', 'proteome_tax_id.tsv')

    output_folder = 'output_folder'
    grouping = 'tax_rank'
    expected_terms = ['1.1.1.86', '2.8.4.1']
    if not os.path.exists(output_folder):
        os.mkdir(output_folder)
    subprocess.call(['esmecata_gseapy', 'gseapy_enrichr', '-f', input_folder, '-o', output_folder, '--grouping', grouping,
                     '--annot-names', ANNOTATION_NAMES_JSON, '--orsumMinTermSize', '4', '--taxon-id', taxon_id_file])

    result_file = os.path.join(output_folder, 'orsum_output_folder', 'filteredResult-Summary.tsv')
    df = pd.read_csv(result_file, sep='\t')
    representing_terms = df['Representing term id']

    assert sorted(representing_terms) == sorted(expected_terms)
    shutil.rmtree(output_folder)


def test_taxon_rank_annotation_enrichment_selected_cli_input_file_picrust2():
    input_folder = os.path.join('test_data', 'picrust_test_predicted_func.tsv')

    output_folder = 'output_folder'
    grouping = 'selected'
    taxa_list_file = os.path.join('test_data', 'taxa_list.tsv')
    expected_terms = ['3.4.21.19', 'ko:K07785']
    if not os.path.exists(output_folder):
        os.mkdir(output_folder)
    subprocess.call(['esmecata_gseapy', 'gseapy_enrichr', '-f', input_folder, '-o', output_folder, '--grouping', grouping, '--taxa-list', taxa_list_file,
                     '--annot-names', ANNOTATION_NAMES_JSON, '--orsumMinTermSize', '4'])

    result_file = os.path.join(output_folder, 'orsum_output_folder', 'filteredResult-Summary.tsv')
    df = pd.read_csv(result_file, sep='\t')
    representing_terms = df['Representing term id']

    assert sorted(representing_terms) == sorted(expected_terms)
    shutil.rmtree(output_folder)


def test_taxon_rank_annotation_enrichment_tax_rank_cli_input_file_picrust2():
    input_annot_file = os.path.join('test_data', 'picrust_test_predicted_func.tsv')
    taxon_id_file = os.path.join('test_data', 'picrust_taxon_id.tsv')

    output_folder = 'output_folder'
    grouping = 'tax_rank'

    expected_terms = ['3.4.21.19', 'ko:K07785']
    if not os.path.exists(output_folder):
        os.mkdir(output_folder)
    subprocess.call(['esmecata_gseapy', 'gseapy_enrichr', '-f', input_annot_file, '-o', output_folder, '--grouping', grouping, '--taxon-id', taxon_id_file,
                     '--annot-names', ANNOTATION_NAMES_JSON, '--orsumMinTermSize', '4'])

    result_file = os.path.join(output_folder, 'orsum_output_folder', 'filteredResult-Summary.tsv')
    df = pd.read_csv(result_file, sep='\t')
    representing_terms = df['Representing term id']

    assert sorted(representing_terms) == sorted(expected_terms)
    shutil.rmtree(output_folder)


def test_taxon_rank_annotation_enrichment_function_selected():
    input_folder = os.path.join('test_data', 'annotation_output')
    output_folder = 'output_folder'
    grouping = 'selected_function'
    function_list_file = os.path.join('test_data', 'function_list.tsv')
    expected_terms = ['org_06', 'org_01']
    if not os.path.exists(output_folder):
        os.mkdir(output_folder)
    taxon_rank_annotation_enrichment(input_folder, output_folder, grouping, function_lists_file=function_list_file, annot_names_file=ANNOTATION_NAMES_JSON,
                                     orsum_minterm_size=4, selected_adjust_pvalue_cutoff=1)

    result_file = os.path.join(output_folder, 'orsum_output_folder', 'filteredResult-Summary.tsv')
    df = pd.read_csv(result_file, sep='\t')
    representing_terms = df['Representing term id']

    assert sorted(representing_terms) == sorted(expected_terms)
    shutil.rmtree(output_folder)


def test_taxon_rank_annotation_enrichment_function_selected_cli():
    input_folder = os.path.join('test_data', 'annotation_output')
    output_folder = 'output_folder'
    grouping = 'selected_function'
    function_list_file = os.path.join('test_data', 'function_list.tsv')
    expected_terms = ['org_06', 'org_01']
    if not os.path.exists(output_folder):
        os.mkdir(output_folder)
    subprocess.call(['esmecata_gseapy', 'gseapy_enrichr', '-f', input_folder, '-o', output_folder, '--grouping', grouping, '--function-list', function_list_file,
                     '--annot-names', ANNOTATION_NAMES_JSON, '--orsumMinTermSize', '4'])

    taxon_rank_annotation_enrichment(input_folder, output_folder, grouping, function_lists_file=function_list_file, annot_names_file=ANNOTATION_NAMES_JSON,
                                     orsum_minterm_size=4, selected_adjust_pvalue_cutoff=1)

    result_file = os.path.join(output_folder, 'orsum_output_folder', 'filteredResult-Summary.tsv')
    df = pd.read_csv(result_file, sep='\t')
    representing_terms = df['Representing term id']

    assert sorted(representing_terms) == sorted(expected_terms)
    shutil.rmtree(output_folder)


def test_taxon_rank_annotation_enrichment_function_selected_picrust2():
    input_annot_file = os.path.join('test_data', 'picrust_test_predicted_func.tsv')
    output_folder = 'output_folder'
    grouping = 'selected_function'
    function_list_file = os.path.join('test_data', 'picrust_function_list.tsv')
    expected_terms = ['org_06', 'org_01']
    if not os.path.exists(output_folder):
        os.mkdir(output_folder)
    taxon_rank_annotation_enrichment(input_annot_file, output_folder, grouping, function_lists_file=function_list_file, annot_names_file=ANNOTATION_NAMES_JSON,
                                     orsum_minterm_size=4, selected_adjust_pvalue_cutoff=0.9)

    result_file = os.path.join(output_folder, 'orsum_output_folder', 'filteredResult-Summary.tsv')
    df = pd.read_csv(result_file, sep='\t')
    representing_terms = df['Representing term id']

    assert sorted(representing_terms) == sorted(expected_terms)
    shutil.rmtree(output_folder)


if __name__ == "__main__":
    test_taxon_rank_annotation_enrichment_selected_cli_input_file_picrust2()



# Copyright (C) 2023-2026 Arnaud Belcour - Univ. Grenoble Alpes, Inria, Microcosme
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

import csv
import os
import sys
import pandas as pd
import numpy as np
import datetime
import time
import json
import pronto
import logging
import subprocess
import urllib.request
import gseapy

from bioservices import KEGG
from Bio.ExPASy import Enzyme
from matplotlib import __version__ as matplotlib_version
from esmecata.core.proteomes import get_taxon_obs_name
from esmecata.utils import get_domain_or_superkingdom_from_ncbi_tax_database
from esmecata import __version__ as esmecata_version
from Bio import __version__ as biopython_version

logger = logging.getLogger(__name__)


def extract_annotation(annotation_reference_folder):
    """ From esmecata annotation reference folder, creates a dict indicating the file associated with each annotation.

    Args:
        annotation_reference_folder (str): path to annotation reference folder
        dataset_annotation_file_path (str): path to output dataset annotation file

    Returns:
        annotation_sets (dict): annotation dict: annotation as key and taxon name as value
    """
    annotation_sets = {}
    for annotation_file in os.listdir(annotation_reference_folder):
        annotation_file_name = os.path.splitext(annotation_file)[0]
        annotation_file_path = os.path.join(annotation_reference_folder, annotation_file)

        with open(annotation_file_path, 'r') as open_annotation_input_file_path:
            csvreader = csv.DictReader(open_annotation_input_file_path, delimiter='\t')

            for line in csvreader:
                gos = line['GO'].split(',')
                ecs = line['EC'].split(',')

                if gos != ['']:
                    for go in gos:
                        if go not in annotation_sets:
                            annotation_sets[go] = [annotation_file_name]
                        else:
                            annotation_sets[go].append(annotation_file_name)
                if ecs != ['']:
                    for ec in ecs:
                        if ec not in annotation_sets:
                            annotation_sets[ec] = [annotation_file_name]
                        else:
                            annotation_sets[ec].append(annotation_file_name)

    return annotation_sets

def extract_observation_name_annotation(annotation_reference_folder):
    """ From esmecata annotation reference folder, creates a dict indicating the file associated with each annotation.

    Args:
        annotation_reference_folder (str): path to annotation reference folder
        dataset_annotation_file_path (str): path to output dataset annotation file

    Returns:
        taxa_annotations_sets (dict): taxon name as key and annotation as value
    """
    taxa_annotations_sets = {}
    for annotation_file in os.listdir(annotation_reference_folder):
        annotation_file_name = os.path.splitext(annotation_file)[0]
        annotation_file_path = os.path.join(annotation_reference_folder, annotation_file)

        with open(annotation_file_path, 'r') as open_annotation_input_file_path:
            csvreader = csv.DictReader(open_annotation_input_file_path, delimiter='\t')

            for line in csvreader:
                gos = line['GO'].split(',')
                ecs = line['EC'].split(',')

                if gos != ['']:
                    for go in gos:
                        if annotation_file_name not in taxa_annotations_sets:
                            taxa_annotations_sets[annotation_file_name] = [go]
                        else:
                            taxa_annotations_sets[annotation_file_name].append(go)
                if ecs != ['']:
                    for ec in ecs:
                        if annotation_file_name not in taxa_annotations_sets:
                            taxa_annotations_sets[annotation_file_name] = [ec]
                        else:
                            taxa_annotations_sets[annotation_file_name].append(ec)

    return taxa_annotations_sets


def read_function_table(function_table_path, transpose_table=True):
    """ Read function table and create dictionary compatible with enrichment analysis.
    The table contains function/annotation as columns and organism as rows.

    Args:
        function_table_path (str): path to function table file.
        transpose_table (bool): boolean

    Returns:
        annotation_sets (dict): if transpose_table is True: annotation as key and list of organism possessing them as values, if transpose_table is False: organism name as key and list of functions possessing them as values,
    """
    df_function_table = pd.read_csv(function_table_path, sep='\t|,', index_col=0, engine='python')

    if transpose_table is True:
        df_function_table = df_function_table.T

    annotation_sets = {}
    for index, row in df_function_table.iterrows():
        annotation_sets[index] = [col for col in df_function_table.columns
                    if row[col]>0
                    for index_iter in range(int(row[col]))]

    return annotation_sets


def get_annot_name(enzyme_datfile, gobasic_file):
    """ Extract annotation names from enzyme.dat and go-baisc.obo.

    Args:
        enzyme_datfile (str): path to enzyme.Dat file
        gobasic_file (str): path to go-baisc.obo file

    Returns:
        enzyme_names (dict): EC as key and EC name as value
        go_names (dict): GO term ID as key and GO term name as value
    """
    enzyme_names = {}
    with open(enzyme_datfile) as infile:
        enzymes = Enzyme.parse(infile)
        for enzyme in enzymes:
            enzyme_id = enzyme['ID']
            enzyme_name = enzyme['DE']
            enzyme_names[enzyme_id] = enzyme_name

    go_ontology = pronto.Ontology(gobasic_file)
    go_names = {}
    for go_term in go_ontology:
        if 'GO:' in go_term:
            go = go_ontology[go_term]
            go_names[go_term] = go.name

    return enzyme_names, go_names


def get_ko_names(ko_file):
    """ Extract KEGG Orthologs names from KEGG database.

    Args:
        ko_file (str): path to KEGG orthologs file

    Returns:
        ko_names (dict): KO as key and KO name as value
    """
    if os.path.exists(ko_file):
        ko_df = pd.read_csv(ko_file, sep='\t')
        ko_df.set_index('KO', inplace=True)
        ko_names = ko_df['name'].to_dict()
    else:
        k = KEGG()
        ko_names = {}
        ko_answer = k.list('ko')
        for ko_line in ko_answer.split('\n'):
            ko_dat = ko_line.split('\t')
            ko_entry = 'ko:'+ko_dat[0]
            if ko_entry.startswith('ko:K'):
                if ';' in ko_dat[1]:
                    ko_name = ko_dat[1].split(';')[1]
                else:
                    ko_name = ko_dat[1]
                ko_names[ko_entry] = ko_name
        ko_df = pd.DataFrame(ko_names.items(), columns=['KO', 'name'])
        ko_df.to_csv(ko_file, sep='\t', index=None)

    return ko_names


def run_orsum(annotation_file_gmt_file, enrichr_module_phylum_output, orsum_output_folder, orsum_minterm_size=None):
    """ Run orsum to reduce list of enriched annotations from gseapy.

    Args:
        annotation_file_gmt_file (str): GMT file containing annotations and their labels
        enrichr_module_phylum_output (str): path to output of gseapy containing enriched annotations for each taxon
        orsum_output_folder (str): path to output folder for orsum
        orsum_minterm_size (int): option minTermSize of orsum
    """
    input_files = [os.path.join(enrichr_module_phylum_output, file) for file in os.listdir(enrichr_module_phylum_output)]
    orsum_cmd = ['orsum.py', '--gmt', annotation_file_gmt_file, '--files', *input_files, '--outputFolder', orsum_output_folder]

    if orsum_minterm_size is not None:
        orsum_cmd.append('--minTermSize')
        orsum_cmd.append(str(orsum_minterm_size))

    logger.info('|EsMeCaTa|gseapy_enrichr| orsum command: {0}'.format(' '.join(orsum_cmd)))
    subprocess.call(orsum_cmd)


def get_orsum_version():
    """Returns version of orsum.

    Returns:
        orsum_version (str): version of orsum.
    """
    orsum_version = None
    response = subprocess.Popen(['orsum.py', '-v'], stdout=subprocess.PIPE, start_new_session=True, universal_newlines="")
    for orsum_line in response.stdout:
        orsum_line_decoded = orsum_line.decode('utf-8')
        orsum_version = orsum_line_decoded.strip('\n')

    if orsum_version is None:
        logger.critical('|EsMeCaTa|gseapy_enrichr| esmecata could not find the version of orsum.')
        logger.critical('|EsMeCaTa|gseapy_enrichr| It is possibly an issue with the installation of orsum (maybe it is not in the PATH). Or it can be due to a change in the output of orsum.py -v command.')
        sys.exit()

    return orsum_version


def extract_organisms_selected(taxa_lists_file):
    """ Extract observation names from input file (each column corresponds to a set of observation name to search for enrichment).

    Args:
        taxa_lists_file (str): path to taxa list file

    Returns:
        taxa_lists (dict): dictionary containing as key the column name of the file and as value the list of observation names
    """
    taxa_lists = {}
    df = pd.read_csv(taxa_lists_file, sep='\t')
    df = df.replace(np.nan, '')
    for col in df.columns:
        taxa_lists[col] = [observation_name for observation_name in df[col].tolist() if observation_name != '']
    return taxa_lists


def taxon_rank_annotation_enrichment(annotation_folder_or_file, output_folder, grouping="tax_rank",
                                     taxon_rank='phylum', taxa_lists_file=None, function_lists_file=None,
                                     annot_names_file='download', orsum_minterm_size=None, selected_adjust_pvalue_cutoff=0.05,
                                     taxon_id_file=None):
    """ Run an enrichment analysis on taxon from annotation results of esmecata using gseapy.
    Then filter this list with orsum.

    Args:
        annotation_folder_or_file (str): path to esmecata annotation folder or function table (from esmecata or other methods)
        output_folder (str): path to output folder
        grouping (str): grouping factor, either "tax_rank", "selected" or "selected_function"
        taxon_rank (str): taxon rank to cluster the observation name together (by default, phylum) when selecting grouping "tax_rank"
        taxa_lists_file (str): path to manually created groups of observation names when selecting grouping "selected"
        function_lists_file (str): path to manually created groups of functions (GO or EC) when selecting grouping "selected_function"
        annot_names_file (str): path to json annot names file. If 'download' is given, download EC number, GO Terms and KO names and generate a json file
        orsum_minterm_size (int): option minTermSize of orsum
        selected_adjust_pvalue_cutoff (float): adjust-Pval cutoff for gseapy enrichr, default: 0.05
        taxon_id_file (str): path to taxon ID file if annotation_folder_or_file is a file
    """
    starttime = time.time()
    logger.info('|EsMeCaTa|gseapy_enrichr| Begin enrichment analysis.')

    if grouping == "tax_rank":
        if os.path.isfile(annotation_folder_or_file) and taxon_id_file is None:
            logger.critical('|EsMeCaTa|gseapy_enrichr| If you give an input annotation file for tax_rank analysis, you have to give a taxon_id_file indicating taxon ID associated with organisms.')
            sys.exit()
        domain_superkingdom_tax_rank_name = get_domain_or_superkingdom_from_ncbi_tax_database()
        if taxon_rank is None:
            logger.critical('|EsMeCaTa|gseapy_enrichr| You have to specify a taxon rank for this analysis with --taxon-rank')
            sys.exit()
        taxon_ranks = ['species', 'genus', 'family', 'order', 'class', 'phylum', 'kingdom', 'domain', 'superkingdom']
        if taxon_rank != 'phylum':
            if taxon_rank not in taxon_ranks:
                logger.critical('|EsMeCaTa|gseapy_enrichr| Incorrect taxon given {0}, possible ranks are: {1}'.format(taxon_rank, ','.join(taxon_ranks)))
                sys.exit()
        if taxon_rank == 'superkingdom':
            if taxon_rank != domain_superkingdom_tax_rank_name and domain_superkingdom_tax_rank_name == 'domain':
                logger.critical('|EsMeCaTa|gseapy_enrichr| superkingdom has been renamed domain in new version of NCBI Taxonomy database, EsMeCaTa will use domain.')
                taxon_rank = domain_superkingdom_tax_rank_name
        if taxon_rank == 'domain':
            if taxon_rank != domain_superkingdom_tax_rank_name and domain_superkingdom_tax_rank_name == 'superkingdom':
                logger.critical('|EsMeCaTa|gseapy_enrichr| domain was named superkingdom in old version of NCBI Taxonomy database (which is the case of your version), EsMeCaTa will use superkingdom.')
                taxon_rank = domain_superkingdom_tax_rank_name

    elif grouping == "selected":
        if taxa_lists_file is None:
            logger.critical('|EsMeCaTa|gseapy_enrichr| You have to specify a taxa lists file for this analysis with --taxa-list.')
            sys.exit()
    elif grouping == "selected_function":
        if function_lists_file is None:
            logger.critical('|EsMeCaTa|gseapy_enrichr| You have to specify a function lists file for this analysis with --function-list.')
            sys.exit()
    elif grouping is None:
        logger.critical('|EsMeCaTa|gseapy_enrichr| You have to choose a grouping factor either "tax_rank", "selected" or "selected_function.')
        sys.exit()

    # Get metadata associated with run.
    options = {'annotation_folder_or_file': annotation_folder_or_file, 'output_folder': output_folder, 'grouping': grouping,
               'taxon_rank': taxon_rank, 'taxa_lists_file': taxa_lists_file,
               'annot_names_file': annot_names_file, 'orsum_minterm_size': orsum_minterm_size,
               'taxon_id_file': taxon_id_file}

    options['tool_dependencies'] = {}
    options['tool_dependencies']['python_package'] = {}
    options['tool_dependencies']['python_package']['Python_version'] = sys.version
    options['tool_dependencies']['python_package']['esmecata'] = esmecata_version
    options['tool_dependencies']['python_package']['pronto'] = pronto.__version__
    options['tool_dependencies']['python_package']['gseapy'] = gseapy.__version__
    options['tool_dependencies']['python_package']['orsum'] = get_orsum_version()
    options['tool_dependencies']['python_package']['pandas'] = pd.__version__
    options['tool_dependencies']['python_package']['biopython'] = biopython_version
    options['tool_dependencies']['python_package']['matplotlib'] = matplotlib_version

    esmecata_metadata = {}
    date = datetime.datetime.now().strftime('%d-%B-%Y %H:%M:%S')
    esmecata_metadata['access_time'] = date
    esmecata_metadata['tool_options'] = options

    if annot_names_file == 'download':
        annot_names = {}
        # Download enzyme.dat.
        enzyme_data_file = os.path.join(output_folder, 'enzyme.dat')
        urllib.request.urlretrieve('https://ftp.expasy.org/databases/enzyme/enzyme.dat', enzyme_data_file)
        # Download go-basic.obo.
        go_basic_obo_file = os.path.join(output_folder, 'go-basic.obo')
        urllib.request.urlretrieve('https://purl.obolibrary.org/obo/go/go-basic.obo', go_basic_obo_file)
        # Retrieves KO name from KEGG.
        ko_file = os.path.join(output_folder, 'ko_names.tsv')
        ko_names = get_ko_names(ko_file)
        annot_names.update(ko_names)

        enzyme_names, go_names = get_annot_name(enzyme_data_file, go_basic_obo_file)
        annot_names.update(enzyme_names)
        annot_names.update(go_names)
        annotation_names_file = os.path.join(output_folder, 'annotation_names.json')
        with open(annotation_names_file, 'w') as open_annotation_names_file:
            json.dump(annot_names, open_annotation_names_file, indent=4)
    else:
        if os.path.exists(annot_names_file):
            with open(annot_names_file, 'r') as open_annotation_names_file:
                annot_names = json.load(open_annotation_names_file)
        else:
            logger.critical('|EsMeCaTa|gseapy_enrichr| --annot-names/annot_names_file must be a valid file or "download" to either use your own annotation names or EC/GO/KO names.')
            sys.exit()

    if os.path.isdir(annotation_folder_or_file):
        proteome_tax_id_file_path = os.path.join(annotation_folder_or_file, 'proteome_tax_id.tsv')

    # Extract the groups that will be search.
    # Either: - taxonomic groups -> tax_rank.
    #  - manually selected groups of organisms -> selected.
    #  - manually selected groups of functions -> selected_function.
    if grouping == "tax_rank":
        if os.path.isfile(annotation_folder_or_file):
            proteome_tax_id_file_path = taxon_id_file
        identified_groups = get_taxon_obs_name(proteome_tax_id_file_path, taxon_rank)
    elif grouping == "selected":
        identified_groups = extract_organisms_selected(taxa_lists_file)
    elif grouping == "selected_function":
        identified_groups = extract_organisms_selected(function_lists_file)
        obs_names = {}
        if os.path.isdir(annotation_folder_or_file):
            with open(proteome_tax_id_file_path, 'r') as proteome_tax_file:
                csvreader = csv.DictReader(proteome_tax_file, delimiter='\t')
                for line in csvreader:
                    observation_name = line['observation_name']
                    tax_name = line['name']
                    obs_names[observation_name] = tax_name

    # Retrieve element that will be searched for over-representation.
    # If input is esmecata annotation folder, iterates on annotation file to generate element sets.
    if os.path.isdir(annotation_folder_or_file):
        annotation_reference_path = os.path.join(annotation_folder_or_file, 'annotation_reference')
        if grouping in ["tax_rank", "selected"]:
            element_sets = extract_annotation(annotation_reference_path)
        elif grouping in ["selected_function"]:
            element_sets = extract_observation_name_annotation(annotation_reference_path)
    # If input is a function table, read it and extract annotations/organisms.
    elif os.path.isfile(annotation_folder_or_file):
        if grouping in ["tax_rank", "selected"]:
            element_sets = read_function_table(annotation_folder_or_file)
        elif grouping in ["selected_function"]:
            element_sets = read_function_table(annotation_folder_or_file, transpose_table=False)

    # Create GMT file for orsum.
    # When using tax_rank or selected parameters, the GMT file contains each annotation, their labels and the observation name in which they were found.
    # When using selected_function, the GMT file contains each taxon, their associated taxon name and the annotations they are linked to. 
    annotation_file_gmt_file = os.path.join(output_folder, 'annotation_file.gmt')
    with open(annotation_file_gmt_file, 'w') as open_output_file:
        csvwriter = csv.writer(open_output_file, delimiter='\t')
        for element in element_sets:
            element_name = 'na'
            if grouping in ["tax_rank", "selected"]:
                if element in annot_names:
                    element_name = element + ' ' + annot_names[element]
                else:
                    element_name = element
            elif grouping in ["selected_function"]:
                if element in obs_names:
                    element_name = element + ' ' + obs_names[element]
                else:
                    element_name = element
            csvwriter.writerow([element, element_name, *element_sets[element]])

    output_dir = os.path.join(output_folder, 'enrichr_module')
    orsum_input_folder = os.path.join(output_folder, 'orsum_input_folder')

    if not os.path.exists(output_dir):
        os.mkdir(output_dir)
    if not os.path.exists(orsum_input_folder):
        os.mkdir(orsum_input_folder)

    enriched_elements = {}
    for group_name in identified_groups:
        identified_group = identified_groups[group_name]
        # Try to run gseapy enrichr, if no enriched results, continue.
        try:
            gseapy.enrichr(gene_list=identified_group, gene_sets=element_sets, background=None,
                        outdir=os.path.join(output_dir, group_name), cutoff=selected_adjust_pvalue_cutoff)
        except ValueError as error:
            logger.critical('|EsMeCaTa|gseapy_enrichr| No enrichred functions with p-value cutoff < {0} for {1}.'.format(selected_adjust_pvalue_cutoff, group_name))
            continue
        if os.path.exists(os.path.join(output_dir, group_name, 'Enrichr.human.enrichr.reports.pdf')):
            # If enriched results, extract the ones with an adjusted p-value inferior to selected_adjust_pvalue_cutoff to output folder (default 0.05).
            df = pd.read_csv(os.path.join(output_dir, group_name, 'Enrichr.human.enrichr.reports.txt'), sep='\t')
            df = df[df['Adjusted P-value'] < selected_adjust_pvalue_cutoff]
            enriched_elements[group_name] = df.set_index('Term')['Adjusted P-value'].to_dict()
            df.sort_values('Adjusted P-value', inplace=True)
            df = df['Term']
            df.to_csv(os.path.join(orsum_input_folder, group_name), sep='\t', index=False, header=False)

    all_elements = set([element for org in enriched_elements for element in enriched_elements[org]])
    all_groups = list(identified_groups.keys())

    enrich_matrix_file = os.path.join(output_folder, 'enrich_matrix.tsv')
    with open(enrich_matrix_file, 'w') as open_output_file:
        csvwriter = csv.writer(open_output_file, delimiter='\t')
        csvwriter.writerow(['Element', 'Name', *all_groups])
        for element in all_elements:
            if grouping in ["tax_rank", "selected"]:
                if element in annot_names:
                    element_name = element + ' ' + annot_names[element]
                else:
                    element_name = 'na'
            elif grouping in ["selected_function"]:
                if element in obs_names:
                    element_name = element + ' ' + obs_names[element]
                else:
                    element_name = 'na'
            csvwriter.writerow([element, element_name, *[enriched_elements[group_name][element] if element in enriched_elements[group_name] else 'NA'  for group_name in all_groups]])

    # Check that there are files in orsum folder.
    orsum_input_folder_files = os.listdir(orsum_input_folder)
    if len(orsum_input_folder_files) == 0:
        logger.critical('|EsMeCaTa|gseapy_enrichr| No enriched files from gseapy enrichr as input for orsum. It seems that there are no enrichd terms found.')
        sys.exit(1)

    # Run orsum to filter list of enriched annotations.
    logger.info('|EsMeCaTa|gseapy_enrichr| Launch orsum visualisation.')
    orsum_output_folder = os.path.join(output_folder, 'orsum_output_folder')
    run_orsum(annotation_file_gmt_file, orsum_input_folder, orsum_output_folder, orsum_minterm_size)

    endtime = time.time()
    duration = endtime - starttime
    logger.info('|EsMeCaTa|gseapy_enrichr| Enrichment analysis took {0}.'.format(duration))

    esmecata_metadata['esmecata_gseapy_enrichr_duration'] = duration
    gseapy_enrichr_metadata_file = os.path.join(output_folder, 'esmecata_metadata_gseapy_enrichr.json')
    if os.path.exists(gseapy_enrichr_metadata_file):
        metadata_files = [metadata_file for metadata_file in os.listdir(output_folder) if 'esmecata_metadata_gseapy_enrichr' in metadata_file]
        gseapy_enrichr_metadata_file = os.path.join(output_folder, 'esmecata_metadata_gseapy_enrichr{0}.json'.format(len(metadata_files)))
        with open(gseapy_enrichr_metadata_file, 'w') as ouput_file:
            json.dump(esmecata_metadata, ouput_file, indent=4)
    else:
        with open(gseapy_enrichr_metadata_file, 'w') as ouput_file:
            json.dump(esmecata_metadata, ouput_file, indent=4)



# Copyright (C) 2024-2026 Arnaud Belcour - Inria, Univ Rennes, CNRS, IRISA Dyliss
# Univ. Grenoble Alpes, Inria, Microcosme
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

import csv
import datetime
import json
import logging
import io
import os
import pandas as pd
import shutil
import sys
import time
import zipfile

from io import TextIOWrapper

from ete4 import __version__ as ete_version
from ete4 import NCBITaxa
from Bio import SeqIO
from Bio import __version__ as biopython_version

from esmecata.utils import is_valid_dir, get_domain_or_superkingdom_from_ncbi_tax_database
from esmecata.core.proteomes import associate_taxon_to_taxon_id, disambiguate_taxon, filter_rank_limit, create_comp_taxonomy_file, taxon_id_to_taxonomic_affiliation
from esmecata.core.eggnog import compute_stat_annotation, write_pathologic
from esmecata.core.clustering import get_proteomes_tax_id_name, compute_openness_pan_proteomes
from esmecata.core.annotation import create_dataset_annotation_file

from esmecata import __version__ as esmecata_version

logger = logging.getLogger(__name__)


def get_taxon_database(archive):
    """ Extract taxon ID contained in precomputed database.

    Args:
        proteome_tax_id_file (str): open file object of esmecata precomputed database

    Returns:
        database_taxon_ids (dict): list of taxon IDs contained in the database
        proteomes_tax_id_names (dict): dict of observation name as key and tax_id_name as value
        taxon_data (dict): dict of observation name as key and tax_id, tax_name, tax_rank and associated proteomes as value
    """
    database_taxon_ids = {}
    proteomes_tax_id_names = {}
    taxon_data = {}
    with archive.open('proteome_tax_id.tsv', 'r') as open_database_taxon_file_path:
        csvreader = csv.DictReader(TextIOWrapper(open_database_taxon_file_path), delimiter='\t')
        for line in csvreader:
            database_taxon_ids[line['tax_id']] = line['name']
            proteomes_tax_id_names[line['tax_id']] = line['tax_id_name']
            taxon_data[line['tax_id']] = [line['tax_id'], line['name'], line['tax_rank'], line['proteome']]

    return database_taxon_ids, proteomes_tax_id_names, taxon_data


def find_proteomes_tax_ids_in_precomputed_database(json_taxonomic_affiliations, database_taxon_ids):
    """Find proteomes associated with taxonomic affiliations

    Args:
        json_taxonomic_affiliations (dict): observation name and dictionary with mapping between taxon name and taxon ID (with remove rank specified)
        database_taxon_ids (dict): list of taxon IDs contained in the database

    Returns:
        association_taxon_database (dict): observation name (key) associated with tax_name and tax_id
        observation_name_not_founds (list): list of observation name which taxonomic affiliation has no match in esmecata database
    """
    logger.info('|EsMeCaTa|precomputed| Search for observation names present in precomputed database.')

    association_taxon_database = {}
    for observation_name in json_taxonomic_affiliations:
        for tax_name in reversed(json_taxonomic_affiliations[observation_name]):
            tax_id = str(json_taxonomic_affiliations[observation_name][tax_name][0])

            # If tax_id has already been found use the corresponding proteomes without new requests.
            if tax_id in database_taxon_ids:
                if tax_name != database_taxon_ids[tax_id]:
                    logger.info('|EsMeCaTa|precomputed| "%s" has a matching taxon ID in the database (taxon ID: %s), but taxon names of input file and the one from database do not match. Taxon name from database will be used: "%s" instead of "%s".', observation_name, tax_id, database_taxon_ids[tax_id], tax_name)
                    tax_name = database_taxon_ids[tax_id]

                else:
                    logger.info('|EsMeCaTa|precomputed| "%s" has a matching taxon ID in the database (taxon ID: %s), it will be associated with the taxon "%s".', observation_name, tax_id, tax_name)

                association_taxon_database[observation_name] = (tax_name,  tax_id)
                break

    observation_name_not_founds = []
    for observation_name in json_taxonomic_affiliations:
        if observation_name not in association_taxon_database:
            logger.info('|EsMeCaTa|precomputed| No taxon available in the database has been found for %s. Maybe, try a new run of esmecata with it?', observation_name)
            observation_name_not_founds.append(observation_name)

    return association_taxon_database, observation_name_not_founds


def search_taxon_database(json_taxonomic_affiliations, database_taxon_file_path, clust_threshold,
                          computed_threshold_folder, annotation_reference_output_folder, reference_proteins_consensus_fasta_folder,
                          pathologic_folder):
    """ Search in esmecata precomputed database from taxa.

    Args:
        json_taxonomic_affiliations (dict): observation name and dictionary with mapping between taxon name and taxon ID (with remove rank specified).
        database_taxon_file_path (str): path to esmecata precomputed database.
        clust_threshold (float): threshold to select protein cluster according to the representation of protein proteome in the cluster (must be equal or superior to the one use in the creation of the precomputed db).
        computed_threshold_folder (str): path to output computed threshold folder.
        annotation_reference_output_folder (str): path to output annotation reference folder.
        reference_proteins_consensus_fasta_folder (str): path to output reference proteins consensus folder.
        pathologic_folder (str): path to output pathologic folder.

    Returns:
        association_taxon_database (dict): observation name (key) associated with tax_name and tax_id.
        proteomes_tax_id_names (dict): dict of observation name as key and tax_id_name as value.
        taxon_data (dict): dict of observation name as key and tax_id, tax_name, tax_rank and associated proteomes as value.
        observation_name_not_founds (list): list of observation name which taxonomic affiliation has no match in esmecata database.
        only_reference_proteome_used (dict): esmecata selected taxon as key and boolean as value indicating if proteomes are reference proteomes.
        proteomes_data_json (dict): esmecata metadta dictionary for proteomes step.
        json_data (dict): esmecata metadata dictionary from database.
        precomputed_db_version (str): version of precomputed database.
    """
    archive = zipfile.ZipFile(database_taxon_file_path, 'r')
    database_taxon_ids, proteomes_tax_id_names, taxon_data = get_taxon_database(archive)

    with archive.open('esmecata_database_metadata.json', 'r') as open_metadata_json:
        json_data = json.load(open_metadata_json)
    for json_key in json_data:
        if json_key.endswith('_proteomes'):
            proteomes_data_json = json_data[json_key]
        if json_key.endswith('_clustering'):
            clustering_data_json = json_data[json_key]

    precomputed_db_version = json_data['database_version']
    logger.critical('|EsMeCaTa|precomputed| EsMeCaTa is using precomputed database "{0}" version {1}.'.format(database_taxon_file_path, precomputed_db_version))

    # Check compatibility between user threshold and database threshold.
    database_clustering_threhsold = clustering_data_json['tool_options']['clust_threshold']
    if clust_threshold < database_clustering_threhsold:
        logger.critical('|EsMeCaTa|precomputed| Selected threshold (-t) too low ({0}) compared to database threhsold ({1}), select one superior or equal to the one of the database.'.format(clust_threshold, database_clustering_threhsold))
        sys.exit(1)

    association_taxon_database, observation_name_not_founds = find_proteomes_tax_ids_in_precomputed_database(json_taxonomic_affiliations, database_taxon_ids)

    # Get only_reference_proteome_used.
    only_reference_proteome_used = {}
    with archive.open('stat_number_proteome.tsv') as open_stat_proteome_file:
        csvreader = csv.DictReader(TextIOWrapper(open_stat_proteome_file), delimiter='\t')
        for line in csvreader:
            only_reference_proteome_used[line['EsMeCaTa_used_taxon']] = line['only_reference_proteome_used']

    tax_id_obs_names = {}
    for observation_name in association_taxon_database:
        tax_id = association_taxon_database[observation_name][1]
        if tax_id != 'not_found':
            if tax_id not in tax_id_obs_names:
                tax_id_obs_names[tax_id] = [observation_name]
            else:
                tax_id_obs_names[tax_id].append(observation_name)

    # For each line of the input files that has a match in the database, recreate an imitation of esmecata output folder.
    for tax_id in tax_id_obs_names:
        tax_id_name = proteomes_tax_id_names[tax_id]

        # Read annotation file
        annotation_file = os.path.join(tax_id_name, tax_id_name+'.tsv')
        with archive.open(annotation_file) as zf:
            df_annotation = pd.read_csv(zf, sep='\t')

        # Create a computed threhsold file.
        output_computed_threshold_file = os.path.join(computed_threshold_folder, tax_id_name+'.tsv')
        df_annotation[['representative_protein', 'cluster_ratio', 'proteomes']].to_csv(output_computed_threshold_file, sep='\t', index=None)

        # Filter according to selected clust_threshold.
        df_annotation = df_annotation[df_annotation['cluster_ratio'] >= clust_threshold]

        kept_protein_ids = set(df_annotation['representative_protein'].tolist())
        for index, observation_name in enumerate(tax_id_obs_names[tax_id]):
            # If there are multiple observation names with the same tax_id, use the first to generate annotation and pathologic files and then copy them.
            if index == 0:
                # Create an annotaiton_reference file for the observation name.
                output_path_annotation_file = os.path.join(annotation_reference_output_folder, observation_name+'.tsv')
                if 'KEGG_reaction' in df_annotation.columns:
                    selected_df_annotation = df_annotation[['representative_protein', 'cluster_members', 'gene_name', 'GO', 'EC', 'KEGG_reaction']]
                    selected_df_annotation.columns = ['protein_cluster', 'cluster_members', 'gene_name', 'GO', 'EC', 'KEGG_reaction']
                else:
                    selected_df_annotation = df_annotation[['representative_protein', 'cluster_members', 'gene_name', 'GO', 'EC']]
                    selected_df_annotation.columns = ['protein_cluster', 'cluster_members', 'gene_name', 'GO', 'EC']
                if not os.path.exists(output_path_annotation_file):
                    selected_df_annotation.to_csv(output_path_annotation_file, sep='\t', index=None)

                # Create pathologic files.
                pathologic_organism_folder = os.path.join(pathologic_folder, observation_name)
                # Add _1 to pathologic file as genetic element cannot have the same name as the organism.
                pathologic_file = os.path.join(pathologic_organism_folder, observation_name+'_1.pf')

                if not os.path.exists(pathologic_file):
                    # Replace NA by empty string.
                    selected_df_annotation = selected_df_annotation.fillna('')
                    reference_proteins_string = selected_df_annotation.set_index('protein_cluster')['cluster_members'].to_dict()
                    reference_proteins = {protein: reference_proteins_string[protein].split(',') for protein in reference_proteins_string}

                    sub_annotated_proteins = selected_df_annotation.set_index('protein_cluster').to_dict('index')
                    # Rename dictionary keys according to the ones expected in write_pathologic.
                    sub_annotated_proteins = [(protein, {'Preferred_name': sub_annotated_proteins[protein]['gene_name'],
                                                        'GOs': sub_annotated_proteins[protein]['GO'],
                                                        'EC': sub_annotated_proteins[protein]['EC']})
                                                        for protein in sub_annotated_proteins]
                    if len(sub_annotated_proteins) > 0:
                        is_valid_dir(pathologic_organism_folder)
                        write_pathologic(observation_name, sub_annotated_proteins, pathologic_file, reference_proteins)
                reference_path_annotation_file = output_path_annotation_file
                reference_pathologic_file = pathologic_file
            else:
                # Copy annotation file.
                output_path_annotation_file = os.path.join(annotation_reference_output_folder, observation_name+'.tsv')
                shutil.copyfile(reference_path_annotation_file, output_path_annotation_file)

                if os.path.exists(reference_pathologic_file):
                    # Create pathologic files.
                    pathologic_organism_folder = os.path.join(pathologic_folder, observation_name)
                    # Add _1 to pathologic file as genetic element cannot have the same name as the organism.
                    pathologic_file = os.path.join(pathologic_organism_folder, observation_name+'_1.pf')

                    # Copy pathologic file.
                    is_valid_dir(pathologic_organism_folder)
                    shutil.copyfile(reference_pathologic_file, pathologic_file)


        # Create a consensus proteomes file.
        clustering_consensus_file = os.path.join(tax_id_name, tax_id_name+'.faa')
        output_path_consensus_file = os.path.join(reference_proteins_consensus_fasta_folder, tax_id_name+'.faa')
        if not os.path.exists(output_path_consensus_file):
            records = []
            with archive.open(clustering_consensus_file) as zf:
                open_zf_text  = io.TextIOWrapper(zf)
                for record in SeqIO.parse(open_zf_text, 'fasta'):
                    if '|' in record.id:
                        protein_id = record.id.split('|')[1]
                    else:
                        protein_id = record.id
                    if protein_id in kept_protein_ids:
                        records.append(record)

            if len(records) > 0:
                logger.critical('|EsMeCaTa|precomputed| {0} protein clusters kept for taxon {1} using threshold {2}.'.format(len(records), tax_id_name, clust_threshold))
                SeqIO.write(records, output_path_consensus_file, 'fasta')
            else:
                logger.critical('|EsMeCaTa|precomputed| 0 protein clusters kept for taxon {0} using threshold {1}, it will not have predictions.'.format(tax_id_name, clust_threshold))

    archive.close()

    return association_taxon_database, proteomes_tax_id_names, taxon_data, observation_name_not_founds, only_reference_proteome_used, proteomes_data_json, json_data, precomputed_db_version


def precomputed_parse_affiliation(input_file, database_taxon_file_path, output_folder, rank_limit=None, update_affiliations=None,
                                  clust_threshold=0.5):
    """From a tsv file with taxonomic affiliations find the associated proteomes and download them.

    Args:
        input_file (str): pathname to the tsv input file containing taxonomic affiliations.
        output_folder (str): pathname to the output folder.
        rank_limit (str): rank limit to filter the affiliations (keep this rank and all inferior ranks).
        update_affiliations (str): option to update taxonomic affiliations.
        clust_threshold (float): threshold to select protein cluster according to the representation of protein proteome in the cluster (must be equal or superior to the one use in the creation of the precomputed db).
    """
    starttime = time.time()
    logger.info('|EsMeCaTa|precomputed| Reading input file.')

    # Check how domain/superkingdom are called in NCBI Taxonomy database.
    domain_superkingdom_tax_rank_name = get_domain_or_superkingdom_from_ncbi_tax_database()
    if rank_limit is not None:
        if rank_limit == 'superkingdom' and rank_limit != domain_superkingdom_tax_rank_name and domain_superkingdom_tax_rank_name == 'domain':
            logger.info('|EsMeCaTa|proteomes| Rank limit sets as superkingdom but it is called domain inside NCBI Taxonomy database, use domain instead.')
            rank_limit = 'domain'
        elif rank_limit == 'domain' and rank_limit != domain_superkingdom_tax_rank_name and domain_superkingdom_tax_rank_name == 'superkingdom':
            logger.info('|EsMeCaTa|proteomes| Rank limit sets as domain but it is called superkingdom inside NCBI Taxonomy database, use superkingdom instead.')
            rank_limit = 'superkingdom'

    is_valid_dir(output_folder)

    proteomes_output_folder = os.path.join(output_folder, '0_proteomes')
    is_valid_dir(proteomes_output_folder)

    clustering_output_folder = os.path.join(output_folder, '1_clustering')
    is_valid_dir(clustering_output_folder)

    computed_threshold_folder = os.path.join(clustering_output_folder, 'computed_threshold')
    is_valid_dir(computed_threshold_folder)

    reference_proteins_consensus_fasta_folder = os.path.join(clustering_output_folder, 'reference_proteins_consensus_fasta')
    is_valid_dir(reference_proteins_consensus_fasta_folder)

    annotation_output_folder = os.path.join(output_folder, '2_annotation')
    is_valid_dir(annotation_output_folder)

    annotation_reference_output_folder = os.path.join(annotation_output_folder, 'annotation_reference')
    is_valid_dir(annotation_reference_output_folder)

    pathologic_folder = os.path.join(annotation_output_folder, 'pathologic')
    is_valid_dir(pathologic_folder)

    # Metadata of the script.
    options = {'input_file': input_file, 'output_folder': output_folder, 'database_taxon_file_path': database_taxon_file_path,
               'rank_limit': rank_limit, 'update_affiliations': update_affiliations, 'clust_threshold': clust_threshold}

    options['tool_dependencies'] = {}
    options['tool_dependencies']['python_package'] = {}
    options['tool_dependencies']['python_package']['Python_version'] = sys.version
    options['tool_dependencies']['python_package']['esmecata'] = esmecata_version
    options['tool_dependencies']['python_package']['ete4'] = ete_version
    options['tool_dependencies']['python_package']['pandas'] = pd.__version__
    options['tool_dependencies']['python_package']['biopython'] = biopython_version

    esmecata_metadata = {}
    date = datetime.datetime.now().strftime('%d-%B-%Y %H:%M:%S')
    esmecata_metadata['access_time'] = date
    esmecata_metadata['tool_options'] = options
    esmecata_metadata['precomputed_database'] = {}

    known_extensions = ['.xlsx', '.tsv', '.csv']
    file_name, file_extension = os.path.splitext(input_file)

    if file_extension not in known_extensions:
        logger.critical('|EsMeCaTa|precomputed| ERROR: Extension ({0}) of input file ({1}) is not part of the compatible extensions ({2}).'.format(file_extension, input_file, ','.join(known_extensions)))
        sys.exit(1)

    if '.xlsx' in input_file:
        df = pd.read_excel(input_file)
    if '.tsv' in input_file:
        df = pd.read_csv(input_file, sep='\t')
    if '.csv' in input_file:
        df = pd.read_csv(input_file, sep=',')

    # Set index on the column containing the OTU name.
    df.set_index('observation_name', inplace=True)

    # taxonomic_affiliation is the column containing the taxonomic affiliation separated by ';': phylum;class;order;family;genus;genus + species
    if 'taxonomic_affiliation' in df.columns:
        taxonomies = df.to_dict()['taxonomic_affiliation']
    else:
        if 'ncbi_taxid' in df.columns:
            ncbi_taxon_ids = df.to_dict()['ncbi_taxid']
        else:
            logger.critical('|EsMeCaTa|proteomes| Missing required columns either: taxonomic_affiliation or ncbi_taxid.')
            sys.exit(1)

    ncbi = NCBITaxa()

    # Parse taxonomic affiliations with ete4 to find matching taxon name.
    if 'taxonomic_affiliation' in df.columns:
        tax_id_names, json_taxonomic_affiliations = associate_taxon_to_taxon_id(taxonomies, update_affiliations, ncbi, output_folder)
    else:
        if 'ncbi_taxid' in df.columns:
            tax_id_names, json_taxonomic_affiliations = taxon_id_to_taxonomic_affiliation(ncbi_taxon_ids, update_affiliations, ncbi, output_folder)

    # Disambiguate taxon name using other taxon from the taxonomic affiliations.
    json_taxonomic_affiliations = disambiguate_taxon(json_taxonomic_affiliations, ncbi)

    # If a rank limit option has been used, limit the taxonomic ranks used.
    if rank_limit:
        json_taxonomic_affiliations = filter_rank_limit(json_taxonomic_affiliations, ncbi, rank_limit)

    json_log = os.path.join(proteomes_output_folder, 'association_taxon_taxID.json')
    with open(json_log, 'w') as ouput_file:
        json.dump(json_taxonomic_affiliations, ouput_file, indent=4)

    predictions_from_which_database = {}
    # If multiple database files are given, iter on each of them to associate taxa with observation name. 
    if ' ' in database_taxon_file_path:
        association_taxon_database = {}
        proteomes_tax_id_names = {}
        taxon_data = {}
        observation_name_not_founds =  []
        only_reference_proteome_used = {}

        json_taxonomic_affiliations_to_research = json_taxonomic_affiliations.copy()
        for single_database_taxon_file_path in database_taxon_file_path.split(' '):
            tmp_database_name = os.path.splitext(os.path.basename(single_database_taxon_file_path))[0]
            tmp_association_taxon_database, tmp_proteomes_tax_id_names, tmp_taxon_data, \
                tmp_observation_name_not_founds, tmp_only_reference_proteome_used, tmp_proteomes_data_json, tmp_json_data, \
                tmp_precomputed_db_version = search_taxon_database(json_taxonomic_affiliations_to_research, single_database_taxon_file_path, clust_threshold,
                                                            computed_threshold_folder, annotation_reference_output_folder, reference_proteins_consensus_fasta_folder,
                                                            pathologic_folder)

            # Extract metadata.
            esmecata_metadata['precomputed_database'][tmp_database_name] = {}
            esmecata_metadata['precomputed_database'][tmp_database_name]['esmecata_query_system'] = tmp_proteomes_data_json['esmecata_query_system']
            esmecata_metadata['precomputed_database'][tmp_database_name]['uniprot_release'] = tmp_proteomes_data_json['uniprot_release']
            esmecata_metadata['precomputed_database'][tmp_database_name]['access_time'] = tmp_proteomes_data_json['access_time']
            esmecata_metadata['precomputed_database'][tmp_database_name]['swissprot_release_number'] = tmp_proteomes_data_json['swissprot_release_number']
            esmecata_metadata['precomputed_database'][tmp_database_name]['swissprot_release_date'] = tmp_proteomes_data_json['swissprot_release_date']
            esmecata_metadata['precomputed_database'][tmp_database_name]['trembl_release_number'] = tmp_proteomes_data_json['trembl_release_number']
            esmecata_metadata['precomputed_database'][tmp_database_name]['trembl_release_date'] = tmp_proteomes_data_json['trembl_release_date']
            esmecata_metadata['precomputed_database'][tmp_database_name]['esmecata_precomputed_db_version'] = tmp_precomputed_db_version

            # Merge all results from the different databases into variables.
            for observation_name in tmp_association_taxon_database:
                predictions_from_which_database[observation_name] = tmp_database_name
                # Remove the observation_name from the ones to be searched with the next database.
                del json_taxonomic_affiliations_to_research[observation_name]
                if observation_name not in association_taxon_database:
                    association_taxon_database[observation_name] = tmp_association_taxon_database[observation_name]

            for tax_id in tmp_proteomes_tax_id_names:
                if tax_id not in proteomes_tax_id_names:
                    proteomes_tax_id_names[tax_id] = tmp_proteomes_tax_id_names[tax_id]

            for tax_id in tmp_taxon_data:
                if tax_id not in taxon_data:
                    taxon_data[tax_id] = tmp_taxon_data[tax_id]

            for observation_name in tmp_observation_name_not_founds:
                if observation_name not in observation_name_not_founds:
                    observation_name_not_founds.append(observation_name)

            for esmecata_taxon in tmp_only_reference_proteome_used:
                if esmecata_taxon not in only_reference_proteome_used:
                    only_reference_proteome_used[esmecata_taxon] = tmp_only_reference_proteome_used[esmecata_taxon]

            # Create metadata file for each database.
            for json_key in tmp_json_data:
                if json_key.endswith('_clustering'):
                    clustering_json_data = tmp_json_data[json_key]
            precomputed_metadata_file = os.path.join(clustering_output_folder, 'esmecata_metadata_clustering_'+tmp_database_name+'.json')
            with open(precomputed_metadata_file, 'w') as ouput_file:
                json.dump(clustering_json_data, ouput_file, indent=4)

            for json_key in tmp_json_data:
                if json_key.endswith('_annotation'):
                    annotation_json_data = tmp_json_data[json_key]
            precomputed_metadata_file = os.path.join(annotation_output_folder, 'esmecata_metadata_annotation_'+tmp_database_name+'.json')
            with open(precomputed_metadata_file, 'w') as ouput_file:
                json.dump(annotation_json_data, ouput_file, indent=4)

        # Ensure that observation names were not found with another database.
        observation_name_not_founds = [observation_name for observation_name in observation_name_not_founds if observation_name not in association_taxon_database]

    else:
        # Run esmecata precomputed on one database.
        association_taxon_database, proteomes_tax_id_names, taxon_data, \
            observation_name_not_founds, only_reference_proteome_used, proteomes_data_json, json_data, \
            precomputed_db_version = search_taxon_database(json_taxonomic_affiliations, database_taxon_file_path, clust_threshold,
                                                        computed_threshold_folder, annotation_reference_output_folder, reference_proteins_consensus_fasta_folder,
                                                        pathologic_folder)

        database_name = os.path.splitext(os.path.basename(database_taxon_file_path))[0]
        for observation_name in association_taxon_database:
            tax_id = association_taxon_database[observation_name][1]
            if tax_id != 'not_found':
                predictions_from_which_database[observation_name] = database_name

        esmecata_metadata['precomputed_database']['esmecata_query_system'] = proteomes_data_json['esmecata_query_system']
        esmecata_metadata['precomputed_database']['uniprot_release'] = proteomes_data_json['uniprot_release']
        esmecata_metadata['precomputed_database']['access_time'] = proteomes_data_json['access_time']
        esmecata_metadata['precomputed_database']['swissprot_release_number'] = proteomes_data_json['swissprot_release_number']
        esmecata_metadata['precomputed_database']['swissprot_release_date'] = proteomes_data_json['swissprot_release_date']
        esmecata_metadata['precomputed_database']['trembl_release_number'] = proteomes_data_json['trembl_release_number']
        esmecata_metadata['precomputed_database']['trembl_release_date'] = proteomes_data_json['trembl_release_date']
        esmecata_metadata['precomputed_database']['esmecata_precomputed_db_version'] = precomputed_db_version

        for json_key in json_data:
            if json_key.endswith('_clustering'):
                clustering_json_data = json_data[json_key]
        precomputed_metadata_file = os.path.join(clustering_output_folder, 'esmecata_metadata_clustering.json')
        with open(precomputed_metadata_file, 'w') as ouput_file:
            json.dump(clustering_json_data, ouput_file, indent=4)

        for json_key in json_data:
            if json_key.endswith('_annotation'):
                annotation_json_data = json_data[json_key]
        precomputed_metadata_file = os.path.join(annotation_output_folder, 'esmecata_metadata_annotation.json')
        with open(precomputed_metadata_file, 'w') as ouput_file:
            json.dump(annotation_json_data, ouput_file, indent=4)

    # Create proteome_tax_id file.
    proteome_tax_id_file = os.path.join(proteomes_output_folder, 'proteome_tax_id.tsv')
    tax_id_obs_names = {}
    proteomes_ids = {}
    with open(proteome_tax_id_file, 'w') as out_file:
        csvwriter = csv.writer(out_file, delimiter='\t')
        csvwriter.writerow(['observation_name', 'name', 'tax_id', 'tax_id_name', 'tax_rank', 'proteome'])
        for observation_name in association_taxon_database:
            tax_id = association_taxon_database[observation_name][1]
            if tax_id != 'not_found':
                tax_name = association_taxon_database[observation_name][0]
                tax_id_name = proteomes_tax_id_names[tax_id]
                tax_rank = taxon_data[tax_id][2]
                proteome = taxon_data[tax_id][3]
                proteomes_ids[observation_name] = (tax_id, proteome.split(','))
                if tax_id not in tax_id_obs_names:
                    tax_id_obs_names[tax_id] = [observation_name]
                else:
                    tax_id_obs_names[tax_id].append(observation_name)
                csvwriter.writerow([observation_name, tax_name, tax_id, tax_id_name, tax_rank, taxon_data[tax_id][3]])

    create_comp_taxonomy_file(json_taxonomic_affiliations, proteomes_ids, tax_id_names, proteomes_output_folder)

    logger.info('|EsMeCaTa|proteomes| Compute proteome stat.')
    # Create stat_proteome.
    run_proteome_tax_id_file = os.path.join(proteomes_output_folder, 'stat_number_proteome.tsv')
    with open(run_proteome_tax_id_file, 'w') as out_file:
        csvwriter = csv.writer(out_file, delimiter='\t')
        csvwriter.writerow(['observation_name', 'Number_proteomes', 'Input_taxon_Name', 'Taxon_rank', 'EsMeCaTa_used_taxon', 'EsMeCaTa_used_rank', 'only_reference_proteome_used'])
        for observation_name in association_taxon_database:
            tax_name = association_taxon_database[observation_name][0]
            tax_id = association_taxon_database[observation_name][1]
            if tax_id != 'not_found':
                tax_rank = taxon_data[tax_id][2]
                nb_tax_proteomes = len(taxon_data[tax_id][3].split(','))
                reversed_affiliation_taxa = list(reversed(list(json_taxonomic_affiliations[observation_name].keys())))
                lowest_tax_rank = None
                for input_tax_name in reversed_affiliation_taxa:
                    if json_taxonomic_affiliations[observation_name][input_tax_name] != ['not_found']:
                        if input_tax_name != 'unknown':
                            lowest_tax_id = json_taxonomic_affiliations[observation_name][input_tax_name][0]
                            lowest_tax_rank = ncbi.get_rank([lowest_tax_id])[lowest_tax_id]
                            break
                csvwriter.writerow([observation_name, nb_tax_proteomes, input_tax_name, lowest_tax_rank, tax_name, tax_rank, only_reference_proteome_used[tax_name]])

    clustering_proteome_tax_id_file = os.path.join(clustering_output_folder, 'proteome_tax_id.tsv')
    shutil.copyfile(proteome_tax_id_file, clustering_proteome_tax_id_file)
    annotation_proteome_tax_id_file = os.path.join(annotation_output_folder, 'proteome_tax_id.tsv')
    shutil.copyfile(proteome_tax_id_file, annotation_proteome_tax_id_file)

    # Create mpwt taxon ID file.
    clustering_taxon_id = {}
    with open(proteome_tax_id_file, 'r') as input_taxon_id_file:
        taxon_id_csvreader = csv.DictReader(input_taxon_id_file, delimiter='\t')
        for line in taxon_id_csvreader:
            clustering_taxon_id[line['observation_name']] = line['tax_id']

    pathologic_taxon_id_file = os.path.join(pathologic_folder, 'taxon_id.tsv')
    with open(pathologic_taxon_id_file, 'w') as taxon_id_file:
        taxon_id_csvwriter = csv.writer(taxon_id_file, delimiter='\t')
        taxon_id_csvwriter.writerow(['species', 'taxon_id'])
        for species in clustering_taxon_id:
            taxon_id_csvwriter.writerow([species, clustering_taxon_id[species]])

    # If some organisms have no match in their taxonomic affiliations with esmecata precomputed database, create an input file for esmecata containing them.
    organism_not_found_in_database_file_path = os.path.join(output_folder, 'organism_not_found_in_database.tsv')
    if len(observation_name_not_founds) > 0:
        df_not_found = df.loc[observation_name_not_founds]
        df_not_found.to_csv(organism_not_found_in_database_file_path, sep='\t')
    else:
        with open(organism_not_found_in_database_file_path, 'w'):
            pass

    logger.info('|EsMeCaTa|proteomes| Generate function table.')
    function_table_file_path = os.path.join(annotation_output_folder, 'function_table.tsv')
    create_dataset_annotation_file(annotation_reference_output_folder, function_table_file_path, 'all')

    logger.info('|EsMeCaTa|proteomes| Compute clustering stat.')
    tax_name_clustering_numbers = {}
    for clustering_file in os.listdir(computed_threshold_folder):
        clustering_file_path = os.path.join(computed_threshold_folder, clustering_file)
        with open(clustering_file_path, 'r') as open_clustering_file_path:
            csvreader = csv.reader(open_clustering_file_path, delimiter='\t')
            next(csvreader)
            cluster_0 = []
            selected_threshold_cluster = []
            cluster_0_95 = []
            for line in csvreader:
                protein_cluster_threshold = float(line[1])
                if protein_cluster_threshold >= 0.95:
                    cluster_0_95.append(line[0])
                if protein_cluster_threshold >= clust_threshold:
                    selected_threshold_cluster.append(line[0])
                if protein_cluster_threshold >= 0:
                    cluster_0.append(line[0])

        if len(selected_threshold_cluster) > 0:
            tax_name_clustering_numbers[clustering_file.replace('.tsv', '')] = [cluster_0, selected_threshold_cluster, cluster_0_95]
    proteomes_taxa_id_names = get_proteomes_tax_id_name(clustering_proteome_tax_id_file)

    clustering_numbers = {}
    for observation_name in proteomes_taxa_id_names:
        tax_name = proteomes_taxa_id_names[observation_name]
        if tax_name in tax_name_clustering_numbers:
            clustering_numbers[observation_name] = tax_name_clustering_numbers[tax_name]

    clustering_stat_file = os.path.join(clustering_output_folder, 'stat_number_clustering.tsv')
    with open(clustering_stat_file, 'w') as stat_file_open:
        csvwriter = csv.writer(stat_file_open, delimiter='\t')
        csvwriter.writerow(['observation_name', 'Number_protein_clusters_panproteome', 'Number_protein_clusters_kept', 'Number_protein_clusters_coreproteome'])
        for observation_name in clustering_numbers:
            cluster_0 = len(clustering_numbers[observation_name][0])
            selected_threshold_cluster = len(clustering_numbers[observation_name][1])
            cluster_0_95 = len(clustering_numbers[observation_name][2])
            csvwriter.writerow([observation_name, cluster_0, selected_threshold_cluster, cluster_0_95])

    logger.info('|EsMeCaTa|proteomes| Compute openness.')
    # Compute openness of proteomes.
    proteome_openness_file = os.path.join(clustering_output_folder, 'stat_openness_proteomes.tsv')
    compute_openness_pan_proteomes(computed_threshold_folder, proteome_openness_file, clustering_threhsold=0.5, iteration_nb=100)

    logger.info('|EsMeCaTa|proteomes| Compute annotation stat.')
    annotation_stat_file = os.path.join(annotation_output_folder, 'stat_number_annotation.tsv')
    compute_stat_annotation(annotation_reference_output_folder, annotation_stat_file)

    endtime = time.time()
    duration = endtime - starttime
    esmecata_metadata['esmecata_precomputed_duration'] = duration
    precomputed_metadata_file = os.path.join(output_folder, 'esmecata_metadata_precomputed.json')
    if os.path.exists(precomputed_metadata_file):
        metadata_files = [metadata_file for metadata_file in os.listdir(output_folder) if 'esmecata_metadata_precomputed' in metadata_file]
        precomputed_metadata_file = os.path.join(output_folder, 'esmecata_metadata_precomputed_{0}.json'.format(len(metadata_files)))
        with open(precomputed_metadata_file, 'w') as ouput_file:
            json.dump(esmecata_metadata, ouput_file, indent=4)
    else:
        with open(precomputed_metadata_file, 'w') as ouput_file:
            json.dump(esmecata_metadata, ouput_file, indent=4)

    stat_precomputed_file = os.path.join(output_folder, 'stat_number_precomputed.tsv')
    df_proteomes = pd.read_csv(run_proteome_tax_id_file, sep='\t')
    df_proteomes.set_index('observation_name', inplace=True)
    df_clustering = pd.read_csv(clustering_stat_file, sep='\t')
    df_clustering.set_index('observation_name', inplace=True)
    df_annotation = pd.read_csv(annotation_stat_file, sep='\t')
    df_annotation.set_index('observation_name', inplace=True)
    df_precomputed_stat = df_proteomes.join([df_clustering, df_annotation])
    df_precomputed_stat['precomputed_db'] = [predictions_from_which_database[obs_name] if obs_name in predictions_from_which_database else '' for obs_name in df_precomputed_stat.index]
    df_precomputed_stat.to_csv(stat_precomputed_file, sep='\t')

    precomputed_metadata_file = os.path.join(proteomes_output_folder, 'esmecata_metadata_proteomes.json')
    with open(precomputed_metadata_file, 'w') as ouput_file:
        json.dump(esmecata_metadata, ouput_file, indent=4)

    logger.info('|EsMeCaTa|precomputed| Extraction of data from database completed in {0}s.'.format(duration))



# Copyright (C) 2023-2026 Alice Mataigne and Pauline Hamon-Giraud - Inria, Univ Rennes, CNRS, IRISA Dyliss
# Arnaud Belcour - Univ. Grenoble Alpes, Inria, Microcosme
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

import os
import json
import logging
import pandas as pd

from esmecata.core.annotation import create_dataset_annotation_file
from esmecata.report.report_creation import create_esmecata_report
from esmecata.report.esmecata2taxontology import get_fig_parameters, generate_sunburst_fig
from esmecata.report.stats_workflow_figures import distributions_by_ranks, compare_ranks_in_out, _format_taxo, create_proteome_representativeness_lineplot_px, \
                                                    data_proteome_representativeness, proteomes_representativeness_details, create_annot_obs_df,annot_frequencies_in_obs, \
                                                    annot_frequencies_in_obs_hist, ec_sunburst
from esmecata.report.esmecata_compression import esmecata_compression_taxonomy_file

logger = logging.getLogger(__name__)


def run_create_workflow_report(input_file, input_folder, output_folder, create_svg=False):
    """ Create report from esmecata workflow output.

    Args:
        input_file (str): pathname to the input file of esmecata.
        input_folder (str): pathname to the output folder of esmecata worklow.
        output_folder (str): output folder path
        create_svg (bool): boolean to create svg files of the figure
    """
    logger.info("--- Launch creation of report for workflow ---")
    create_esmecata_report(input_file, input_folder, output_folder, create_svg)


def run_proteomes_report_creation(input_folder, output_folder, create_svg=False):
    """ Create report from esmecata proteomes output.

    Args:
        input_folder (str): pathname to the output folder of esmecata proteomes.
        output_folder (str): output folder path
        create_svg (bool): boolean to create svg files of the figure
    """
    logger.info("--- Launch creation of report for proteomes ---")
    df_stat = pd.read_csv(os.path.join(input_folder, 'stat_number_proteome.tsv'), header=0, index_col='observation_name', sep='\t')
    df_proteome_tax_id = pd.read_csv(os.path.join(input_folder, 'proteome_tax_id.tsv'), header=0, index_col='observation_name', sep='\t')
    df_taxon_proteome_tax_id = _format_taxo(df_proteome_tax_id)
    df_stat_join = df_stat.join(df_proteome_tax_id)
    df_stat_join = df_stat_join.join(df_taxon_proteome_tax_id)
    fig1 = distributions_by_ranks(df_stat_join, output_folder, 15, 'phylum')

    with open(os.path.join(input_folder, 'association_taxon_taxID.json')) as f:
        association_taxon_tax_id = json.load(f)
    fig4 = compare_ranks_in_out(df_proteome_tax_id, association_taxon_tax_id, output_folder)

    tax_comp_file = os.path.join(input_folder, 'taxonomy_diff.tsv')
    output_file = os.path.join(output_folder, 'esmecata2taxonomy')
    data = get_fig_parameters(tax_comp_file)
    fig = generate_sunburst_fig(data, output_file)

    output_file = os.path.join(output_folder, 'esmecata_compression')
    fig10 = esmecata_compression_taxonomy_file(tax_comp_file, output_file, False)


def run_clustering_report_creation(input_folder, output_folder, create_svg=False):
    """ Create report from esmecata clustering output.

    Args:
        input_folder (str): pathname to the output folder of esmecata clustering.
        output_folder (str): output folder path
        create_svg (bool): boolean to create svg files of the figure
    """
    logger.info("--- Launch creation of report for clustering ---")

    df_proteome_tax_id = pd.read_csv(os.path.join(input_folder, 'proteome_tax_id.tsv'), header=0, index_col='observation_name', sep='\t')
    computed_threshold_folder = os.path.join(input_folder, 'computed_threshold')
    df_clustering = data_proteome_representativeness(df_proteome_tax_id, computed_threshold_folder)

    metadata_clustering = json.load(open(os.path.join(input_folder, 'esmecata_metadata_clustering.json'), 'r'))

    fig12 = create_proteome_representativeness_lineplot_px(df_clustering,
        metadata_clustering["tool_options"]["clust_threshold"],
        output_folder)

    fig12_details = proteomes_representativeness_details(df_clustering,
        metadata_clustering["tool_options"]["clust_threshold"],
        output_folder)


def run_annotation_report_creation(input_folder, output_folder, create_svg=False):
    """ Create report from esmecata annotation output.

    Args:
        input_folder (str): pathname to the output folder of esmecata annotation.
        output_folder (str): output folder path
        create_svg (bool): boolean to create svg files of the figure
    """
    annotation_reference_folder = os.path.join(input_folder, 'annotation_reference')
    dataset_annotation_ec_file = os.path.join(output_folder, 'dataset_annotation_ec.tsv')
    dataset_annotation_go_file = os.path.join(output_folder, 'dataset_annotation_go.tsv')

    _ = create_dataset_annotation_file(annotation_reference_folder, dataset_annotation_ec_file, 'EC')

    _ = create_dataset_annotation_file(annotation_reference_folder, dataset_annotation_go_file, 'GO')

    data_ec_number = create_annot_obs_df(dataset_annotation_ec_file, output_folder, 'EC numbers')

    data_go_term = create_annot_obs_df(dataset_annotation_go_file, output_folder, "GO terms")

    fig5 = annot_frequencies_in_obs(data_ec_number["df_annot_frequencies"], output_folder, "EC numbers")
    #fig6 = fraction_of_all_annot_in_obs(data_ec_number["df_fractionin_obs"], DATA["DF_STATS"], output_folder, "EC numbers")
    fig7 = annot_frequencies_in_obs_hist(data_ec_number["df_annot_frequencies"], output_folder, "EC numbers")
    #fig8 = fraction_of_all_annot_in_obs_hist(data_ec_number["df_fractionin_obs"], DATA["DF_STATS"], output_folder, "EC numbers")

    fig5b = annot_frequencies_in_obs(data_go_term["df_annot_frequencies"], output_folder, "GO terms")
    #fig6b = fraction_of_all_annot_in_obs(data_go_term["df_fractionin_obs"], DATA["DF_STATS"], output_folder, "GO terms")
    fig7b = annot_frequencies_in_obs_hist(data_go_term["df_annot_frequencies"], output_folder, "GO terms")
    #fig8b = fraction_of_all_annot_in_obs_hist(data_go_term["df_fractionin_obs"], DATA["DF_STATS"], output_folder, "GO terms")

    fig11 = ec_sunburst(data_ec_number["df_annot_frequencies"].index, output_folder)


import csv
import os
import shutil
import subprocess

from esmecata.report.workflow_create_report import run_create_workflow_report
from esmecata.core.precomputed import precomputed_parse_affiliation


def test_create_report_from_precomputed_offline():
    output_folder = 'output_folder'
    buchnera_workflow_file = os.path.join('test_data', 'buchnera_workflow.tsv')
    buchnera_database_file = os.path.join('test_data', 'buchnera_database.zip')
    precomputed_parse_affiliation(buchnera_workflow_file, buchnera_database_file, output_folder)

    report_output_folder = 'output_report_folder'
    run_create_workflow_report(buchnera_workflow_file, output_folder, report_output_folder)

    html_report_file = os.path.join(report_output_folder, 'esmecata_summary.html')
    # Check creation of HTML report file.
    assert os.path.exists(html_report_file)


def test_create_report_from_precomputed_cli_offline():
    output_folder = 'output_folder'
    buchnera_workflow_file = os.path.join('test_data', 'buchnera_workflow.tsv')
    buchnera_database_file = os.path.join('test_data', 'buchnera_database.zip')

    subprocess.call(['esmecata', 'precomputed', '-i', buchnera_workflow_file, '-d', buchnera_database_file, '-o', output_folder])

    report_output_folder = 'output_report_folder'
    subprocess.call(['esmecata_report', 'create_report', '-i', buchnera_workflow_file, '-f', output_folder, '-o', report_output_folder])

    html_report_file = os.path.join(report_output_folder, 'esmecata_summary.html')
    # Check creation of HTML report file.
    assert os.path.exists(html_report_file)

if __name__ == "__main__":
    test_create_report_from_precomputed_offline()
    test_create_report_from_precomputed_cli_offline()
